import pathlib
# replace this import when dropping support for Python 3.8
# from importlib import resources as importlib_resources
import importlib_resources
import sys
import tempfile
import time

import h5py
import numpy as np
import pytest

import dclab
from dclab.features import emodulus
from dclab.rtdc_dataset import feat_anc_core

from helper_methods import example_data_dict, retrieve_data


@pytest.mark.filterwarnings('ignore::dclab.features.emodulus.'
                            + 'YoungsModulusLookupTableExceededWarning')
def test_af_emodulus_ancil():
    keys = ["area_um", "deform"]
    ddict = example_data_dict(size=8472, keys=keys)
    # known-media
    ds2 = dclab.new_dataset(ddict)
    ds2.config["setup"]["flow rate"] = 0.16
    ds2.config["setup"]["channel width"] = 30
    ds2.config["imaging"]["pixel size"] = .34
    ds2.config["calculation"] = {"emodulus lut": "LE-2D-FEM-19",
                                 "emodulus viscosity model": "herold-2017",
                                 "emodulus medium": "CellCarrier",
                                 "emodulus temperature": 23.0
                                 }
    emod1 = ds2["emodulus"][:]
    ds2.config["calculation"]["emodulus viscosity model"] = \
        "buyukurganci-2022"
    emod2 = ds2["emodulus"][:]

    assert not np.allclose(emod1, emod2, equal_nan=True)


@pytest.mark.filterwarnings('ignore::dclab.features.emodulus.'
                            + 'YoungsModulusLookupTableExceededWarning')
def test_af_emodulus_known_media():
    keys = ["area_um", "deform"]
    ddict = example_data_dict(size=8472, keys=keys)
    # known-media
    ds2 = dclab.new_dataset(ddict)
    ds2.config["setup"]["flow rate"] = 0.16
    ds2.config["setup"]["channel width"] = 30
    ds2.config["imaging"]["pixel size"] = .34
    ds2.config["calculation"] = {"emodulus lut": "LE-2D-FEM-19",
                                 "emodulus viscosity model": "herold-2017",
                                 "emodulus medium": "CellCarrier",
                                 "emodulus temperature": 23.0
                                 }
    # ancillary feature priority check
    for af in feat_anc_core.AncillaryFeature.get_instances("emodulus"):
        if af.priority % 2 == 0:  # remove this case when removing deprecations
            # exclude the old deprecated ancillary features
            continue
        if af.data == "case C":
            assert af.is_available(ds2)
        else:
            assert not af.is_available(ds2)


def test_af_emodulus_known_media_error_set_viscosity():
    keys = ["area_um", "deform"]
    ddict = example_data_dict(size=8472, keys=keys)
    # legacy
    ds = dclab.new_dataset(ddict)
    ds.config["setup"]["flow rate"] = 0.16
    ds.config["setup"]["channel width"] = 30
    ds.config["imaging"]["pixel size"] = .34
    ds.config["calculation"] = {"emodulus lut": "LE-2D-FEM-19",
                                "emodulus viscosity model": "herold-2017",
                                "emodulus medium": "CellCarrier",
                                "emodulus temperature": 23.0,
                                "emodulus viscosity": 0.5
                                }
    with pytest.raises(ValueError, match="must not"):
        ds.__getitem__("emodulus")


@pytest.mark.skipif(sys.version_info < (3, 3),
                    reason="perf_counter requires python3.3 or higher")
@pytest.mark.filterwarnings('ignore::dclab.features.emodulus.'
                            + 'YoungsModulusLookupTableExceededWarning')
def test_af_emodulus_cache():
    keys = ["area_um", "deform"]
    ddict = example_data_dict(size=8472, keys=keys)
    ds = dclab.new_dataset(ddict)
    ds.config["setup"]["flow rate"] = 0.16
    ds.config["setup"]["channel width"] = 30
    ds.config["imaging"]["pixel size"] = .34
    ds.config["calculation"] = {"emodulus lut": "LE-2D-FEM-19",
                                "emodulus viscosity model": "herold-2017",
                                "emodulus medium": "CellCarrier",
                                "emodulus temperature": 23.0,
                                }
    t1 = time.perf_counter()
    assert "emodulus" in ds
    t2 = time.perf_counter()

    t3 = time.perf_counter()
    ds["emodulus"]
    t4 = time.perf_counter()
    assert t4 - t3 > t2 - t1


def test_af_emodulus_legacy_area():
    # computes "area_um" from "area_cvx"
    keys = ["area_cvx", "deform"]
    ddict = example_data_dict(size=8472, keys=keys)
    ds = dclab.new_dataset(ddict)
    # area can be computed from areapix
    ds.config["imaging"]["pixel size"] = .34
    assert "area_um" in ds
    assert "emodulus" not in ds, "not config for emodulus"
    ds.config["setup"]["flow rate"] = 0.16
    ds.config["setup"]["channel width"] = 30
    ds.config["calculation"] = {"emodulus lut": "LE-2D-FEM-19",
                                "emodulus viscosity model": "herold-2017",
                                "emodulus medium": "CellCarrier",
                                "emodulus temperature": 23.0,
                                "emodulus viscosity": 0.5
                                }
    assert "emodulus" in ds
    with pytest.raises(ValueError, match="must not"):
        ds.__getitem__("emodulus")


def test_af_emodulus_legacy_none():
    keys = ["area_msd", "deform"]
    ddict = example_data_dict(size=8472, keys=keys)
    ds = dclab.new_dataset(ddict)
    assert "emodulus" not in ds, "not config for emodulus"
    ds.config["calculation"] = {"emodulus lut": "LE-2D-FEM-19",
                                "emodulus viscosity model": "herold-2017",
                                "emodulus medium": "CellCarrier",
                                "emodulus temperature": 23.0,
                                }
    assert "emodulus" not in ds, "column 'area_um' should be missing"


def test_af_emodulus_legacy_none2():
    keys = ["area_um", "deform"]
    ddict = example_data_dict(size=8472, keys=keys)
    ds = dclab.new_dataset(ddict)
    assert "emodulus" not in ds, "not config for emodulus"
    ds.config["calculation"] = {"emodulus medium": "CellCarrier",
                                "emodulus viscosity model": "herold-2017",
                                "emodulus temperature": 23.0,
                                "emodulus viscosity": 0.5
                                }
    assert "emodulus" not in ds, "emodulus lut should be missing"


def test_af_emodulus_reservoir():
    """Reservoir measurements should not have emodulus"""
    keys = ["area_um", "deform"]
    ddict = example_data_dict(size=8472, keys=keys)
    # legacy
    ds = dclab.new_dataset(ddict)
    ds.config["setup"]["flow rate"] = 0.16
    ds.config["setup"]["channel width"] = 30
    ds.config["imaging"]["pixel size"] = .34
    ds.config["calculation"] = {"emodulus lut": "LE-2D-FEM-19",
                                "emodulus viscosity model": "herold-2017",
                                "emodulus medium": "CellCarrier",
                                "emodulus temperature": 23.0,
                                }
    assert "emodulus" in ds
    ds2 = dclab.new_dataset(ddict)
    ds2.config["setup"]["flow rate"] = 0.16
    ds2.config["setup"]["channel width"] = 30
    ds2.config["imaging"]["pixel size"] = .34
    ds2.config["calculation"] = {"emodulus lut": "LE-2D-FEM-19",
                                 "emodulus viscosity model": "herold-2017",
                                 "emodulus medium": "CellCarrier",
                                 "emodulus temperature": 23.0,
                                 }
    ds2.config["setup"]["chip region"] = "reservoir"
    assert "emodulus" not in ds2


@pytest.mark.filterwarnings('ignore::dclab.features.emodulus.'
                            + 'YoungsModulusLookupTableExceededWarning')
def test_af_emodulus_temp_feat():
    keys = ["area_um", "deform"]
    ddict = example_data_dict(size=8472, keys=keys)
    # legacy
    ds = dclab.new_dataset(ddict)
    ds.config["setup"]["flow rate"] = 0.16
    ds.config["setup"]["channel width"] = 30
    ds.config["imaging"]["pixel size"] = .34
    ds.config["calculation"] = {"emodulus lut": "LE-2D-FEM-19",
                                "emodulus viscosity model": "herold-2017",
                                "emodulus medium": "CellCarrier",
                                "emodulus temperature": 23.0,
                                }
    ddict2 = example_data_dict(size=8472, keys=keys)
    ddict2["temp"] = 23.0 * np.ones(8472)
    # temp-feat
    ds2 = dclab.new_dataset(ddict2)
    ds2.config["setup"]["flow rate"] = 0.16
    ds2.config["setup"]["channel width"] = 30
    ds2.config["imaging"]["pixel size"] = .34
    ds2.config["calculation"] = {"emodulus lut": "LE-2D-FEM-19",
                                 "emodulus viscosity model": "herold-2017",
                                 "emodulus medium": "CellCarrier",
                                 }
    assert np.sum(~np.isnan(ds["emodulus"])) > 0
    assert np.allclose(ds["emodulus"], ds2["emodulus"], equal_nan=True,
                       rtol=0, atol=1e-7)
    # ancillary feature priority check
    for af in feat_anc_core.AncillaryFeature.get_instances("emodulus"):
        if af.priority % 2 == 0:  # remove this case when removing deprecations
            # exclude the old deprecated ancillary features
            continue
        if af.data == "case A":
            assert af.is_available(ds2)
        else:
            assert not af.is_available(ds2)


@pytest.mark.filterwarnings('ignore::dclab.features.emodulus.'
                            + 'YoungsModulusLookupTableExceededWarning')
def test_af_emodulus_temp_feat_2():
    keys = ["area_um", "deform"]
    ddict = example_data_dict(size=8472, keys=keys)
    # legacy
    ds = dclab.new_dataset(ddict)
    ds.config["setup"]["flow rate"] = 0.16
    ds.config["setup"]["channel width"] = 30
    ds.config["imaging"]["pixel size"] = .34
    ds.config["calculation"] = {"emodulus lut": "LE-2D-FEM-19",
                                "emodulus viscosity model": "herold-2017",
                                "emodulus medium": "CellCarrier",
                                "emodulus temperature": 23.0,
                                }
    ddict2 = example_data_dict(size=8472, keys=keys)
    ddict2["temp"] = 23.0 * np.ones(8472)
    ddict2["temp"][0] = 23.5  # change first element
    # temp-feat
    ds2 = dclab.new_dataset(ddict2)
    ds2.config["setup"]["flow rate"] = 0.16
    ds2.config["setup"]["channel width"] = 30
    ds2.config["imaging"]["pixel size"] = .34
    ds2.config["calculation"] = {"emodulus medium": "CellCarrier",
                                 "emodulus lut": "LE-2D-FEM-19",
                                 "emodulus viscosity model": "herold-2017",
                                 }
    assert np.sum(~np.isnan(ds["emodulus"])) > 0
    assert np.allclose(ds["emodulus"][1:], ds2["emodulus"][1:], equal_nan=True,
                       rtol=0, atol=1e-7)
    assert not np.allclose(ds["emodulus"][0], ds2["emodulus"][0])
    ds3 = dclab.new_dataset(ddict)
    ds3.config["setup"]["flow rate"] = 0.16
    ds3.config["setup"]["channel width"] = 30
    ds3.config["imaging"]["pixel size"] = .34
    ds3.config["calculation"] = {"emodulus lut": "LE-2D-FEM-19",
                                 "emodulus viscosity model": "herold-2017",
                                 "emodulus medium": "CellCarrier",
                                 "emodulus temperature": 23.5,
                                 }
    assert np.allclose(ds3["emodulus"][0], ds2["emodulus"][0], rtol=0,
                       atol=1e-7)


@pytest.mark.filterwarnings('ignore::dclab.features.emodulus.'
                            + 'YoungsModulusLookupTableExceededWarning')
def test_af_emodulus_temp_feat_with_basin():
    """
    In dclab 0.60.0 there was a bug: When the temperature feature was
    in a mapped basin, then array operations did not work and the Young's
    modulus could not be computed
    """
    path = retrieve_data("fmt-hdf5_image-mask-blood_2021.zip")
    path_basin = path.with_name("basin.rtdc")

    # Create a basin for temperature
    with h5py.File(path, "a") as hin, \
            h5py.File(path_basin, "w") as hbn:
        assert "temp" not in hin
        assert len(hin["events/image"]) == 18
        hbn.attrs.update(hin.attrs)
        hbn["events/temp"] = np.linspace(21, 22, 18, endpoint=True)

        hw = dclab.RTDCWriter(hin)
        hw.store_basin(
            basin_name="temperature",
            basin_locs=[path_basin],
            basin_format="hdf5",
            basin_type="file",
        )

    with dclab.new_dataset(path) as ds:

        ds.config["calculation"] = {"emodulus lut": "LE-2D-FEM-19",
                                    "emodulus viscosity model": "herold-2017",
                                    "emodulus medium": "CellCarrier",
                                    }

        assert np.allclose(
            ds["emodulus"],
            np.array([0.88053377, 0.8585773, 0.92474311,
                      np.nan, np.nan, np.nan,
                      np.nan, 0.62051569, 0.85272409,
                      0.74716945, 0.83060064, 0.79674873,
                      0.85180626, 0.87352672, np.nan,
                      0.82130422, 0.82563875, 0.66465965]),
            atol=.2, rtol=0, equal_nan=True
        )


@pytest.mark.filterwarnings('ignore::dclab.features.emodulus.'
                            + 'YoungsModulusLookupTableExceededWarning')
def test_af_emodulus_temp_feat_with_basin_mapped():
    """
    In dclab 0.60.0 there was a bug: When the temperature feature was
    in a mapped basin, then array operations did not work and the Young's
    modulus could not be computed
    """
    path = retrieve_data("fmt-hdf5_image-mask-blood_2021.zip")
    path_basin = path.with_name("basin.rtdc")

    # Create a basin for temperature
    with h5py.File(path, "a") as hin, \
            h5py.File(path_basin, "w") as hbn:
        assert "temp" not in hin
        assert len(hin["events/image"]) == 18
        hbn.attrs.update(hin.attrs)
        temp = np.full(22, 23)
        temp[:18] = np.linspace(21, 22, 18, endpoint=True)
        hbn["events/temp"] = temp

        hw = dclab.RTDCWriter(hin)
        hw.store_basin(
            basin_name="temperature",
            basin_locs=[path_basin],
            basin_format="hdf5",
            basin_type="file",
            basin_map=np.arange(18),
        )

    with dclab.new_dataset(path) as ds:
        ds.config["calculation"] = {"emodulus lut": "LE-2D-FEM-19",
                                    "emodulus viscosity model": "herold-2017",
                                    "emodulus medium": "CellCarrier",
                                    }

        assert np.allclose(
            ds["emodulus"],
            np.array([0.88053377, 0.8585773, 0.92474311,
                      np.nan, np.nan, np.nan,
                      np.nan, 0.62051569, 0.85272409,
                      0.74716945, 0.83060064, 0.79674873,
                      0.85180626, 0.87352672, np.nan,
                      0.82130422, 0.82563875, 0.66465965]),
            atol=.2, rtol=0, equal_nan=True
        )


@pytest.mark.filterwarnings('ignore::dclab.features.emodulus.'
                            + 'YoungsModulusLookupTableExceededWarning')
def test_af_emodulus_temp_feat_with_basin_mapped_internal():
    """
    In dclab 0.60.0 there was a bug: When the temperature feature was
    in a mapped basin, then array operations did not work and the Young's
    modulus could not be computed
    """
    path = retrieve_data("fmt-hdf5_image-mask-blood_2021.zip")
    path_basin = path.with_name("basin.rtdc")

    # Create a basin for temperature
    with h5py.File(path, "a") as hin, \
            h5py.File(path_basin, "w") as hbn:
        assert "temp" not in hin
        assert len(hin["events/image"]) == 18
        hbn.attrs.update(hin.attrs)
        temp = np.full(22, 23)
        temp[1:19] = np.linspace(21, 22, 18, endpoint=True)

        hw = dclab.RTDCWriter(hin)
        hw.store_basin(
            basin_name="temperature",
            basin_format="h5dataset",
            basin_locs=["basin_events"],
            basin_type="internal",
            basin_map=np.arange(1, 19),
            basin_feats=["temp"],
            internal_data={"temp": temp}
        )

    with dclab.new_dataset(path) as ds:
        ds.config["calculation"] = {"emodulus lut": "LE-2D-FEM-19",
                                    "emodulus viscosity model": "herold-2017",
                                    "emodulus medium": "CellCarrier",
                                    }

        assert np.allclose(
            ds["emodulus"],
            np.array([0.88053377, 0.8585773, 0.92474311,
                      np.nan, np.nan, np.nan,
                      np.nan, 0.62051569, 0.85272409,
                      0.74716945, 0.83060064, 0.79674873,
                      0.85180626, 0.87352672, np.nan,
                      0.82130422, 0.82563875, 0.66465965]),
            atol=.2, rtol=0, equal_nan=True
        )


@pytest.mark.filterwarnings('ignore::dclab.features.emodulus.'
                            + 'YoungsModulusLookupTableExceededWarning')
def test_af_emodulus_visc_only():
    keys = ["area_um", "deform"]
    ddict = example_data_dict(size=8472, keys=keys)
    # legacy
    ds = dclab.new_dataset(ddict)
    ds.config["setup"]["flow rate"] = 0.16
    ds.config["setup"]["channel width"] = 30
    ds.config["imaging"]["pixel size"] = .34
    ds.config["calculation"] = {"emodulus lut": "LE-2D-FEM-19",
                                "emodulus viscosity model": "herold-2017",
                                "emodulus medium": "CellCarrier",
                                "emodulus temperature": 23.0,
                                }
    # visc-only
    ds2 = dclab.new_dataset(ddict)
    ds2.config["setup"]["flow rate"] = 0.16
    ds2.config["setup"]["channel width"] = 30
    ds2.config["imaging"]["pixel size"] = .34
    visc = dclab.features.emodulus.viscosity.get_viscosity(
        medium="CellCarrier",
        channel_width=30,
        flow_rate=0.16,
        temperature=23.0,
        model="herold-2017")
    ds2.config["calculation"] = {"emodulus lut": "LE-2D-FEM-19",
                                 "emodulus viscosity": visc
                                 }
    assert np.sum(~np.isnan(ds["emodulus"])) > 0
    assert np.allclose(ds["emodulus"], ds2["emodulus"], equal_nan=True,
                       rtol=0, atol=1e-7)
    # ancillary feature priority check
    for af in feat_anc_core.AncillaryFeature.get_instances("emodulus"):
        if af.priority % 2 == 0:  # remove this case when removing deprecations
            # exclude the old deprecated ancillary features
            continue
        if af.data == "case B":
            assert af.is_available(ds2)
        else:
            assert not af.is_available(ds2)


@pytest.mark.filterwarnings('ignore::dclab.features.emodulus.'
                            + 'YoungsModulusLookupTableExceededWarning')
def test_af_emodulus_visc_only_2():
    keys = ["area_um", "deform"]
    ddict = example_data_dict(size=8472, keys=keys)
    visc = dclab.features.emodulus.viscosity.get_viscosity(
        medium="CellCarrier",
        channel_width=30,
        flow_rate=0.16,
        temperature=23.0,
        model="herold-2017",
    )
    # legacy
    ds = dclab.new_dataset(ddict)
    ds.config["setup"]["flow rate"] = 0.16
    ds.config["setup"]["channel width"] = 30
    ds.config["imaging"]["pixel size"] = .34
    ds.config["calculation"] = {"emodulus lut": "LE-2D-FEM-19",
                                "emodulus medium": "other",
                                "emodulus viscosity": visc
                                }
    # visc-only
    ds2 = dclab.new_dataset(ddict)
    ds2.config["setup"]["flow rate"] = 0.16
    ds2.config["setup"]["channel width"] = 30
    ds2.config["imaging"]["pixel size"] = .34
    ds2.config["calculation"] = {"emodulus lut": "LE-2D-FEM-19",
                                 "emodulus viscosity": visc
                                 }
    assert np.sum(~np.isnan(ds["emodulus"])) > 0
    assert np.allclose(ds["emodulus"], ds2["emodulus"], equal_nan=True,
                       rtol=0, atol=1e-7)


def test_bad_lut_data():
    try:
        emodulus.load_lut("bad_string_asdkubhasd")
    except ValueError:
        pass
    else:
        assert False, "Invalid `lut_data` results in ValueError"


@pytest.mark.filterwarnings(
    'ignore::dclab.features.emodulus.KnowWhatYouAreDoingWarning')
def test_extrapolate():
    """Test whether spline interpolation gives reasonable results"""
    lut, _ = emodulus.load_lut("LE-2D-FEM-19")

    area_norm = lut[:, 0].max()
    emodulus.normalize(lut[:, 0], area_norm)

    deform_norm = lut[:, 1].max()
    emodulus.normalize(lut[:, 1], deform_norm)

    np.random.seed(47)
    more_than_5perc = []
    valid_ones = 0

    for _ in range(100):
        # pick a few values from the LUT
        ids = np.random.randint(0, lut.shape[0], 10)
        area_um = lut[ids, 0]
        deform = lut[ids, 1]
        # set the emodulus to zero
        emod = np.nan * np.zeros(deform.size)
        # "extrapolate" within the grid using the spline
        emodulus.extrapolate_emodulus(
            lut=lut,
            datax=area_um,
            deform=deform,
            emod=emod,
            deform_norm=deform_norm,
            inplace=True)
        valid = ~np.isnan(emod)
        valid_ones += np.sum(valid)
        res = np.abs(lut[ids, 2] - emod)[valid] / lut[ids, 2][valid]
        if np.sum(res > .05):
            more_than_5perc.append([ids, res])

    assert len(more_than_5perc) == 0
    assert valid_ones == 151


def test_load_lut_from_array():
    ref_lut, ref_meta = emodulus.load_lut("LE-2D-FEM-19")
    lut2, meta2 = emodulus.load_lut((ref_lut, ref_meta))
    assert np.all(ref_lut == lut2)
    assert ref_meta == meta2
    assert ref_lut is not lut2, "data should be copied"
    assert ref_meta is not meta2, "meta data should be copied"


def test_load_lut_from_path():
    ref_lut, ref_meta = emodulus.load_lut("LE-2D-FEM-19")
    ref = importlib_resources.files(
        "dclab.features.emodulus") / "lut_LE-2D-FEM-19.txt"
    with importlib_resources.as_file(ref) as path:
        lut2, meta2 = emodulus.load_lut(path)
        assert np.all(ref_lut == lut2)
        assert ref_meta == meta2


def test_load_lut_from_badobject():
    try:
        emodulus.load_lut({"test": "nonesense"})
    except ValueError:
        pass
    else:
        assert False, "dict should not be supported"


def test_load_lut_from_badpath():
    try:
        emodulus.load_lut("peter/pan.txt")
    except ValueError:
        pass
    else:
        assert False, "dict should not be supported"


def test_pixelation_correction_volume():
    ddelt = emodulus.get_pixelation_delta(feat_corr="deform",
                                          feat_absc="volume",
                                          data_absc=100,
                                          px_um=0.34)
    assert np.allclose(ddelt, 0.011464479831134636)


def test_register_external_lut():
    """Load an external LUT and compute YM data"""
    identifier = "test-test_register_external_lut"
    ref = importlib_resources.files(
        "dclab.features.emodulus") / "lut_LE-2D-FEM-19.txt"
    with importlib_resources.as_file(ref) as path:
        emodulus.register_lut(path, identifier=identifier)
        # cleanup
        emodulus.load.EXTERNAL_LUTS.clear()


@pytest.mark.filterwarnings('ignore::dclab.features.emodulus.'
                            + 'YoungsModulusLookupTableExceededWarning')
def test_register_external_lut_and_get_emodulus():
    keys = ["area_um", "deform"]
    ddict = example_data_dict(size=8472, keys=keys)
    # from internal LUT
    ds = dclab.new_dataset(ddict)
    ds.config["setup"]["flow rate"] = 0.16
    ds.config["setup"]["channel width"] = 30
    ds.config["imaging"]["pixel size"] = .34
    ds.config["calculation"] = {"emodulus lut": "LE-2D-FEM-19",
                                "emodulus viscosity model": "herold-2017",
                                "emodulus medium": "CellCarrier",
                                "emodulus temperature": 23.0
                                }
    assert np.sum(~np.isnan(ds["emodulus"])) > 0
    # from external LUT
    identifier = "test-test_register_external_lut"
    ref = importlib_resources.files(
        "dclab.features.emodulus") / "lut_LE-2D-FEM-19.txt"
    with importlib_resources.as_file(ref) as path:
        emodulus.register_lut(path, identifier=identifier)
        ds2 = dclab.new_dataset(ddict)
        ds2.config["setup"]["flow rate"] = 0.16
        ds2.config["setup"]["channel width"] = 30
        ds2.config["imaging"]["pixel size"] = .34
        ds2.config["calculation"] = {"emodulus lut": identifier,
                                     "emodulus viscosity model": "herold-2017",
                                     "emodulus medium": "CellCarrier",
                                     "emodulus temperature": 23.0
                                     }
        assert np.sum(~np.isnan(ds2["emodulus"])) > 0
        assert np.allclose(ds["emodulus"], ds2["emodulus"], equal_nan=True,
                           rtol=0, atol=1e-7)


@pytest.mark.filterwarnings('ignore::dclab.features.emodulus.'
                            + 'YoungsModulusLookupTableExceededWarning')
def test_register_external_lut_with_internal_identifier():
    keys = ["area_um", "deform"]
    ddict = example_data_dict(size=8472, keys=keys)
    # from internal LUT
    ds = dclab.new_dataset(ddict)
    ds.config["setup"]["flow rate"] = 0.16
    ds.config["setup"]["channel width"] = 30
    ds.config["imaging"]["pixel size"] = .34
    ds.config["calculation"] = {"emodulus lut": "LE-2D-FEM-19",
                                "emodulus viscosity model": "herold-2017",
                                "emodulus medium": "CellCarrier",
                                "emodulus temperature": 23.0
                                }
    assert np.sum(~np.isnan(ds["emodulus"])) > 0
    # from external LUT
    ref = importlib_resources.files(
        "dclab.features.emodulus") / "lut_LE-2D-FEM-19.txt"
    with importlib_resources.as_file(ref) as path:
        path2 = pathlib.Path(
            tempfile.mkdtemp("external_lut_with_id")) / "lut.txt"
        text = pathlib.Path(path).read_text().split("\n")
        text[33] = '#   "identifier": "LE-2D-FEM-19b",'
        path2.write_text("\n".join(text))
        emodulus.register_lut(path2)


@pytest.mark.filterwarnings('ignore::dclab.features.emodulus.'
                            + 'YoungsModulusLookupTableExceededWarning')
def test_register_external_lut_without_identifier():
    keys = ["area_um", "deform"]
    ddict = example_data_dict(size=8472, keys=keys)
    # from internal LUT
    ds = dclab.new_dataset(ddict)
    ds.config["setup"]["flow rate"] = 0.16
    ds.config["setup"]["channel width"] = 30
    ds.config["imaging"]["pixel size"] = .34
    ds.config["calculation"] = {"emodulus lut": "LE-2D-FEM-19",
                                "emodulus viscosity model": "herold-2017",
                                "emodulus medium": "CellCarrier",
                                "emodulus temperature": 23.0
                                }
    assert np.sum(~np.isnan(ds["emodulus"])) > 0
    # from external LUT
    ref = importlib_resources.files(
        "dclab.features.emodulus") / "lut_LE-2D-FEM-19.txt"
    with importlib_resources.as_file(ref) as path:
        with pytest.raises(ValueError):
            emodulus.register_lut(path)


@pytest.mark.filterwarnings('ignore::dclab.features.emodulus.'
                            + 'YoungsModulusLookupTableExceededWarning')
def test_simple_emod():
    x = np.linspace(0, 250, 100)
    y = np.linspace(0, 0.1, 100)
    x, y = np.meshgrid(x, y)

    emod = emodulus.get_emodulus(area_um=x,
                                 deform=y,
                                 medium="CellCarrier",
                                 channel_width=30,
                                 flow_rate=0.16,
                                 px_um=0,  # without pixelation correction
                                 temperature=23,
                                 visc_model="herold-2017")

    assert np.allclose(emod[10, 50], 1.1875799054283109)
    assert np.allclose(emod[50, 50], 0.5527066911133949)
    assert np.allclose(emod[80, 50], 0.4567858941760323)

    assert np.allclose(emod[10, 80], 1.5744560306483262)
    assert np.allclose(emod[50, 80], 0.73534561544655519)
    assert np.allclose(emod[80, 80], 0.60737083178222251)



"""Compress .rtdc files"""
from __future__ import annotations

import argparse
import pathlib
import warnings

import hdf5plugin
import h5py

from ..rtdc_dataset import rtdc_copy, RTDCWriter
from .. import util
from .._version import version

from . import common


def compress(
        path_in: str | pathlib.Path = None,
        path_out: str | pathlib.Path = None,
        force: bool = False,
        check_suffix: bool = True,
        ret_path: bool = False,
        ):
    """Create a new dataset with all features compressed lossless

    Parameters
    ----------
    path_in: str or pathlib.Path
        file to compress
    path_out: str or pathlib
        output file path
    force: bool
        DEPRECATED
    check_suffix: bool
        check suffixes for input and output files
    ret_path: bool
        whether to return the output path

    Returns
    -------
    path_out: pathlib.Path (optional)
        output path (with possibly corrected suffix)
    """
    cmp_kw = hdf5plugin.Zstd(clevel=5)
    if path_out is None or path_in is None:
        parser = compress_parser()
        args = parser.parse_args()
        path_in = args.input
        path_out = args.output
        force = args.force

    allowed_input_suffixes = [".rtdc"]
    if not check_suffix:
        allowed_input_suffixes.append(pathlib.Path(path_in).suffix)

    path_in, path_out, path_temp = common.setup_task_paths(
        path_in, path_out, allowed_input_suffixes=allowed_input_suffixes)

    if force:
        warnings.warn(
            "The `force` keyword argument is deprecated since dclab 0.49.0, "
            "because compressed HDF5 Datasets are now copied and there "
            "is no reason to avoid or use force anymore.",
            DeprecationWarning)

    # command log
    logs = {"dclab-compress": common.get_command_log(paths=[path_in])}

    with warnings.catch_warnings(record=True) as w:
        warnings.simplefilter("always")
        with h5py.File(path_in) as h5, h5py.File(path_temp, "w") as hc:
            rtdc_copy(src_h5file=h5,
                      dst_h5file=hc,
                      features="all",
                      include_basins=True,
                      include_logs=True,
                      include_tables=True,
                      meta_prefix="",
                      )

            hc.require_group("logs")
            # rename old dclab-compress logs
            for lkey in ["dclab-compress", "dclab-compress-warnings"]:
                if lkey in hc["logs"]:
                    # This is cached, so no worry calling it multiple times.
                    md55m = util.hashfile(path_in, count=80)
                    # rename
                    hc["logs"][f"{lkey}_{md55m}"] = hc["logs"][lkey]
                    del hc["logs"][lkey]

        # warnings log
        if w:
            logs["dclab-compress-warnings"] = common.assemble_warnings(w)

    # Write log file
    with RTDCWriter(path_temp,
                    compression_kwargs=cmp_kw,
                    mode="append") as hw:
        for name in logs:
            hw.store_log(name, logs[name])

    # Finally, rename temp to out
    path_temp.rename(path_out)

    if ret_path:
        return path_out


def compress_parser():
    descr = "Create a compressed version of an .rtdc file. This can be " \
            + "used for saving disk space (loss-less compression). The " \
            + "data generated during an experiment is usually not compressed."
    parser = argparse.ArgumentParser(description=descr)
    parser.add_argument('input', metavar="INPUT", type=str,
                        help='Input path (.rtdc file)')
    parser.add_argument('output', metavar="OUTPUT", type=str,
                        help='Output path (.rtdc file)')
    parser.add_argument('--force',
                        dest='force',
                        action='store_true',
                        help='DEPRECATED')
    parser.set_defaults(force=False)
    parser.add_argument('--version', action='version',
                        version=f'dclab-compress {version}')
    return parser



"""RT-DC hdf5 format"""
from __future__ import annotations

import io
import json
import pathlib
from typing import Any, BinaryIO, Dict
import warnings

import h5py

from ...external.packaging import parse as parse_version
from ...util import hashobj, hashfile

from ..config import Configuration
from ..core import RTDCBase

from . import events
from . import logs
from . import tables

#: rtdc files exported with dclab prior to this version are not supported
MIN_DCLAB_EXPORT_VERSION = "0.3.3.dev2"


class OldFormatNotSupportedError(BaseException):
    pass


class UnknownKeyWarning(UserWarning):
    pass


class RTDC_HDF5(RTDCBase):
    def __init__(self,
                 h5path: str | pathlib.Path | BinaryIO | io.IOBase,
                 h5kwargs: Dict[str, Any] = None,
                 *args,
                 **kwargs):
        """HDF5 file format for RT-DC measurements

        Parameters
        ----------
        h5path: str or pathlib.Path or file-like object
            Path to an '.rtdc' measurement file or a file-like object
        h5kwargs: dict
            Additional keyword arguments given to :class:`h5py.File`
        *args:
            Arguments for `RTDCBase`
        **kwargs:
            Keyword arguments for `RTDCBase`

        Attributes
        ----------
        path: pathlib.Path
            Path to the experimental HDF5 (.rtdc) file
        """
        super(RTDC_HDF5, self).__init__(*args, **kwargs)

        # Any subclass from RTDC_HDF5 is probably a remote-type and should
        # not be able to access local basins. If you do not agree, please
        # enable this in the definition of the subclass.
        self._local_basins_allowed = True if self.format == "hdf5" else False

        if isinstance(h5path, (str, pathlib.Path)):
            h5path = pathlib.Path(h5path)
        else:
            h5path = h5path

        self._hash = None
        self.path = h5path

        # Increase the read cache (which defaults to 1MiB), since
        # normally we have around 2.5MiB image chunks.
        if h5kwargs is None:
            h5kwargs = {}
        h5kwargs.setdefault("rdcc_nbytes", 10 * 1024 ** 2)
        h5kwargs.setdefault("rdcc_w0", 0)

        self.h5kwargs = h5kwargs
        self.h5file = h5py.File(h5path, **h5kwargs)

        self._events = events.H5Events(self.h5file)

        # Parse configuration
        self.config = RTDC_HDF5.parse_config(self.h5file)

        # Override logs property with HDF5 data
        self.logs = logs.H5Logs(self.h5file)

        # Override the tables property with HDF5 data
        self.tables = tables.H5Tables(self.h5file)

        # check version
        rtdc_soft = self.config["setup"].get("software version", "unknown")
        if rtdc_soft.startswith("dclab "):
            rtdc_ver = parse_version(rtdc_soft.split(" ")[1])
            if rtdc_ver < parse_version(MIN_DCLAB_EXPORT_VERSION):
                msg = "The file {} was created ".format(self.path) \
                      + "with dclab {} which is ".format(rtdc_ver) \
                      + "not supported anymore! Please rerun " \
                      + "dclab-tdms2rtdc / export the data again."
                raise OldFormatNotSupportedError(msg)

        self.title = "{} - M{}".format(
            self.config["experiment"].get("sample", "undefined sample"),
            self.config["experiment"].get("run index", "0"))

    def close(self):
        """Close the underlying HDF5 file"""
        super(RTDC_HDF5, self).close()
        self.h5file.close()

    @property
    def _h5(self):
        warnings.warn("Access to the underlying HDF5 file is now public. "
                      "Please use the `h5file` attribute instead of `_h5`!",
                      DeprecationWarning)
        return self.h5file

    @staticmethod
    def can_open(h5path):
        """Check whether a given file is in the .rtdc file format"""
        h5path = pathlib.Path(h5path)
        if h5path.suffix == ".rtdc":
            return True
        else:
            # we don't know the extension; check for the "events" group
            canopen = False
            try:
                # This is a workaround for Python2 where h5py cannot handle
                # unicode file names.
                with h5path.open("rb") as fd:
                    h5 = h5py.File(fd, "r")
                    if "events" in h5:
                        canopen = True
            except IOError:
                # not an HDF5 file
                pass
            return canopen

    @staticmethod
    def parse_config(h5path):
        """Parse the RT-DC configuration of an HDF5 file

        `h5path` may be a h5py.File object or an actual path
        """
        if not isinstance(h5path, h5py.File):
            with h5py.File(h5path, mode="r") as fh5:
                h5attrs = dict(fh5.attrs)
        else:
            h5attrs = dict(h5path.attrs)

        # Convert byte strings to unicode strings
        # https://github.com/h5py/h5py/issues/379
        for key in h5attrs:
            if isinstance(h5attrs[key], bytes):
                h5attrs[key] = h5attrs[key].decode("utf-8")

        config = Configuration()
        for key in h5attrs:
            section, pname = key.split(":")
            config[section][pname] = h5attrs[key]
        return config

    @property
    def hash(self):
        """Hash value based on file name and content"""
        if self._hash is None:
            tohash = [self.path.name,
                      # Hash a maximum of ~1MB of the hdf5 file
                      hashfile(self.path, blocksize=65536, count=20)]
            self._hash = hashobj(tohash)
        return self._hash

    def basins_get_dicts(self):
        """Return list of dicts for all basins defined in `self.h5file`"""
        return self.basin_get_dicts_from_h5file(self.h5file)

    @staticmethod
    def basin_get_dicts_from_h5file(h5file):
        """Return list of dicts for all basins defined in `h5file`"""
        basins = []
        # Do not sort anything here, sorting is done in `RTDCBase`.
        for bk in h5file.get("basins", []):
            bdat = list(h5file["basins"][bk])
            if isinstance(bdat[0], bytes):
                bdat = [bi.decode("utf") for bi in bdat]
            bdict = json.loads(" ".join(bdat))
            bdict["key"] = bk
            basins.append(bdict)
        return basins



from __future__ import annotations

from collections.abc import Mapping
import copy
import json
import os
import pathlib
from typing import Dict, List, Literal, Tuple
import warnings

import h5py
import hdf5plugin
import numpy as np

from .. import definitions as dfn
from ..util import hashobj
from .._version import version

from .feat_anc_plugin import PlugInFeature
from .meta_table import MetaTable

#: DEPRECATED (use `CHUNK_SIZE_BYTES` instead)
CHUNK_SIZE = 100

#: Chunks size in bytes for storing HDF5 datasets
CHUNK_SIZE_BYTES = 1024**2  # 1MiB

#: features that should be written to the output file as uint32 values
FEATURES_UINT32 = [
    "fl1_max",
    "fl1_npeaks",
    "fl2_max",
    "fl2_npeaks",
    "fl3_max",
    "fl3_npeaks",
    "index",
    "ml_class",
    "nevents",
]

#: features that should be written to the output file as uint64 values
FEATURES_UINT64 = [
    "frame",
]


class RTDCWriter:
    def __init__(self,
                 path_or_h5file: str | pathlib.Path | h5py.File,
                 mode: Literal['append', 'replace', 'reset'] = "append",
                 compression_kwargs: Dict | Mapping = None,
                 compression: str = "deprecated"):
        """RT-DC data writer classe

        Parameters
        ----------
        path_or_h5file: str or pathlib.Path or h5py.Group
            Path to an HDF5 file or an HDF5 file opened in write mode
        mode: str
            Defines how the data are stored:

              - "append": append new feature data to existing h5py Datasets
              - "replace": replace existing h5py Datasets with new features
                (used for ancillary feature storage)
              - "reset": do not keep any previous data
        compression_kwargs: dict-like
            Dictionary with the keys "compression" and "compression_opts"
            which are passed to :func:`h5py.H5File.create_dataset`. The
            default is Zstandard compression with the lowest compression
            level `hdf5plugin.Zstd(clevel=1)`. To disable compression, use
            `{"compression": None}`.
        compression: str or None
            Compression method used for data storage;
            one of [None, "lzf", "gzip", "szip"].

            .. deprecated:: 0.43.0
                Use `compression_kwargs` instead.
        """
        if mode not in ["append", "replace", "reset"]:
            raise ValueError(f"Invalid mode '{mode}'!")
        if compression != "deprecated":
            warnings.warn("The `compression` kwarg is deprecated in favor of "
                          "`compression_kwargs`!",
                          DeprecationWarning)
            if compression_kwargs is not None:
                raise ValueError("You may not specify `compression` and "
                                 "`compression_kwargs` at the same time!")
            # be backwards-compatible
            compression_kwargs = {"compression": compression}
        if compression_kwargs is None:
            compression_kwargs = hdf5plugin.Zstd(clevel=1)

        self.mode = mode
        self.compression_kwargs = compression_kwargs
        if isinstance(path_or_h5file, h5py.Group):
            self.owns_path = False
            self.path = pathlib.Path(path_or_h5file.file.filename)
            self.h5file = path_or_h5file
            if mode == "reset":
                raise ValueError("'reset' mode incompatible with h5py.Group!")
        else:
            self.owns_path = True
            self.path = pathlib.Path(path_or_h5file)
            self.h5file = h5py.File(path_or_h5file,
                                    mode=("w" if mode == "reset" else "a"))
        #: unfortunate necessity, as `len(h5py.Group)` can be really slow
        self._group_sizes = {}

    def __enter__(self):
        return self

    def __exit__(self, type, value, tb):
        # close the HDF5 file
        try:
            self.h5file.require_group("events")
            if len(self.h5file["events"]):
                self.rectify_metadata()
            self.version_brand()
        except BaseException:
            raise
        finally:
            # This is guaranteed to run if any exception is raised.
            self.close()

    @staticmethod
    def get_best_nd_chunks(item_shape, item_dtype=np.float64):
        """Return best chunks for HDF5 datasets

        Chunking has performance implications. It’s recommended to keep the
        total size of dataset chunks between 10 KiB and 1 MiB. This number
        defines the maximum chunk size as well as half the maximum cache
        size for each dataset.
        """
        # Note that `np.prod(()) == 1`
        event_size = np.prod(item_shape) * np.dtype(item_dtype).itemsize

        chunk_size = CHUNK_SIZE_BYTES / event_size
        # Set minimum chunk size to 10 so that we can have at least some
        # compression performance.
        chunk_size_int = max(10, int(np.floor(chunk_size)))
        return tuple([chunk_size_int] + list(item_shape))

    def close(self):
        """Close the underlying HDF5 file if a path was given during init"""
        if self.owns_path:
            self.h5file.close()

    def rectify_metadata(self):
        """Autocomplete the metadta of the RTDC-measurement

        The following configuration keys are updated:

        - experiment:event count
        - fluorescence:samples per event
        - imaging: roi size x (if image or mask is given)
        - imaging: roi size y (if image or mask is given)

        The following configuration keys are added if not present:

        - fluorescence:channel count
        """
        # set event count
        feats = sorted(self.h5file.get("events", {}).keys())
        if feats:
            self.h5file.attrs["experiment:event count"] = len(
                self.h5file["events"][feats[0]])
        else:
            raise ValueError(f"No features in '{self.path}'!")

        # ignore empty features in the checks further below
        for feat in feats[:]:  # iterate over a copy of the list
            obj = self.h5file["events"][feat]
            if ((isinstance(feat, h5py.Dataset) and obj.shape[0] == 0)  # ds
                    or len(obj) == 0):  # groups
                feats.remove(feat)

        # set samples per event
        if "trace" in feats:
            traces = list(self.h5file["events"]["trace"].keys())
            trsize = self.h5file["events"]["trace"][traces[0]].shape[1]
            self.h5file.attrs["fluorescence:samples per event"] = trsize

        # set channel count
        chcount = sum(
            ["fl1_max" in feats, "fl2_max" in feats, "fl3_max" in feats])
        if chcount:
            if "fluorescence:channel count" not in self.h5file.attrs:
                self.h5file.attrs["fluorescence:channel count"] = chcount

        # set roi size x/y
        if "image" in feats:
            shape = self.h5file["events"]["image"][0].shape
        elif "mask" in feats:
            shape = self.h5file["events"]["mask"][0].shape
        else:
            shape = None
        if shape is not None:
            # update shape
            self.h5file.attrs["imaging:roi size x"] = shape[1]
            self.h5file.attrs["imaging:roi size y"] = shape[0]

    def store_basin(self,
                    basin_name: str,
                    basin_type: Literal['file', 'internal', 'remote'],
                    basin_format: str,
                    basin_locs: List[str | pathlib.Path],
                    basin_descr: str | None = None,
                    basin_feats: List[str] = None,
                    basin_map: np.ndarray | Tuple[str, np.ndarray] = None,
                    internal_data: Dict | h5py.Group = None,
                    verify: bool = True,
                    ):
        """Write basin information

        Parameters
        ----------
        basin_name: str
            basin name; Names do not have to be unique.
        basin_type: str
            basin type (file or remote); Files are paths accessible by the
            operating system (including e.g. network shares) whereas
            remote locations normally require an active internet connection.
        basin_format: str
            The basin format must match the ``format`` property of an
            :class:`.RTDCBase` subclass (e.g. "hdf5" or "dcor")
        basin_locs: list
            location of the basin as a string or (optionally)
            a ``pathlib.Path``
        basin_descr: str
            optional string describing the basin
        basin_feats: list of str
            list of features this basin provides; You may use this to
            restrict access to features for a specific basin.
        basin_map: np.ndarray or tuple of (str, np.ndarray)
            If this is an integer numpy array, it defines the mapping
            of event indices from the basin dataset to the referring dataset
            (the dataset being written to disk). Normally, the basinmap
            feature used for storing the mapping information is inferred
            from the currently defined basinmap features. However, if you
            are incepting basins, then this might not be sufficient, and you
            have to specify explicitly which basinmap feature to use. In such
            a case, you may specify a tuple `(feature_name, mapping_array)`
            where `feature_name` is the explicit mapping name, e.g.
            `"basinmap3"`.
        internal_data: dict or instance of h5py.Group
            A dictionary or an `h5py.Group` containing the basin data.
            The data are copied to the "basin_events" group, if
            `internal_data` is not an `h5py.Group` in the current HDF5 file.
            This must be specified when storing internal basins, and it
            must not be specified for any other basin type.
        verify: bool
            whether to verify the basin before storing it; You might have
            set this to False if you would like to write a basin that is
            e.g. temporarily not available

        Returns
        -------
        basin_hash: str
            hash of the basin which serves as the name of the HDF5 dataset
            stored in the output file

            .. versionadded:: 0.58.0
        """
        if basin_type == "internal":
            if internal_data is None:
                raise ValueError(
                    "When writing an internal basin, you must specify "
                    "`internal_data` which is either a dictionary of numpy "
                    "arrays or an `h5py.Group` containing the relevant "
                    "datasets.")
            if (isinstance(internal_data, dict)
                    or (isinstance(internal_data, h5py.Group)
                        and internal_data.file != self.h5file)):
                # The data are not yet stored in this HDF5 file
                for feat in basin_feats:
                    igroup = self.h5file.require_group("basin_events")
                    if feat in igroup:
                        raise ValueError(f"The feature '{feat}' already "
                                         f"exists in the 'basin_events' group")
                    self.write_ndarray(group=igroup,
                                       name=feat,
                                       data=internal_data[feat])
                # just override it with the default
                basin_locs = ["basin_events"]
            elif verify:
                # Verify the existence of the data inside this HDF5 file
                if basin_locs != ["basin_events"]:
                    warnings.warn("You specified an uncommon location for "
                                  f"your internal basins: {basin_locs}. "
                                  f"Please use 'basin_events' instead.")
                for feat in basin_feats:
                    if feat not in self.h5file[basin_locs[0]]:
                        raise ValueError(f"Could not find feature '{feat}' in "
                                         f"the group [{basin_locs[0]}]")

        # Expand optional tuple for basin_map
        if isinstance(basin_map, (list, tuple)) and len(basin_map) == 2:
            basin_map_name, basin_map = basin_map
        else:
            basin_map_name = None

        if verify and basin_type in ["file", "remote"]:
            # We have to import this here to avoid circular imports
            from .load import new_dataset
            # Make sure the basin can be opened by dclab, verify its ID
            cur_id = self.h5file.attrs.get("experiment:run identifier")
            for loc in basin_locs:
                with new_dataset(loc) as ds:
                    # We can open the file, which is great.
                    if cur_id:
                        # Compare the IDs.
                        ds_id = ds.get_measurement_identifier()
                        if not (ds_id == cur_id
                                or (basin_map is not None
                                    and cur_id.startswith(ds_id))):
                            raise ValueError(
                                f"Measurement identifier mismatch between "
                                f"{self.path} ({cur_id}) and {loc} ({ds_id})!")
            if basin_feats:
                for feat in basin_feats:
                    if not dfn.feature_exists(feat):
                        raise ValueError(f"Invalid feature: '{feat}'")
            if basin_map is not None:
                if (not isinstance(basin_map, np.ndarray)
                        and basin_map.dtype != np.uint64):
                    raise ValueError(
                        "The array specified in `basin_map` argument must be "
                        "a numpy array with the dtype `np.uint64`!")

        # determine the basinmap to use
        if basin_map is not None:
            self.h5file.require_group("events")
            if basin_map_name is None:
                # We have to determine the basin_map_name to use for this
                # mapped basin.
                for ii in range(10):  # basinmap0 to basinmap9
                    bm_cand = f"basinmap{ii}"
                    if bm_cand in self.h5file["events"]:
                        # There is a basin mapping defined in the file. Check
                        # whether it is identical to ours.
                        if np.all(self.h5file["events"][bm_cand] == basin_map):
                            # Great, we are done here.
                            basin_map_name = bm_cand
                            break
                        else:
                            # This mapping belongs to a different basin,
                            # try the next mapping.
                            continue
                    else:
                        # The mapping is not defined in the dataset, and we may
                        # write it to a new feature.
                        basin_map_name = bm_cand
                        self.store_feature(feat=basin_map_name, data=basin_map)
                        break
                else:
                    raise ValueError(
                        "You have exhausted the usage of mapped basins for "
                        "the current dataset. Please revise your analysis "
                        "pipeline.")
            else:
                if basin_map_name not in self.h5file["events"]:
                    # Write the explicit basin mapping into the file.
                    self.store_feature(feat=basin_map_name, data=basin_map)
                elif not np.all(
                        self.h5file["events"][basin_map_name] == basin_map):
                    # This is a sanity check that we have to perform.
                    raise ValueError(
                        f"The basin mapping feature {basin_map_name} you "
                        f"specified explicitly already exists in "
                        f"{self.h5file} and they do not match. I assume "
                        f"you are trying explicitly write to a basinmap that "
                        f"is already used elsewhere.")
        else:
            # Classic, simple case
            basin_map_name = "same"

        b_data = {
            "description": basin_descr,
            "format": basin_format,
            "name": basin_name,
            "type": basin_type,
            "features": None if basin_feats is None else sorted(basin_feats),
            "mapping": basin_map_name,
        }
        if basin_type == "file":
            flocs = []
            for pp in basin_locs:
                pp = pathlib.Path(pp)
                if verify:
                    flocs.append(str(pp.resolve()))
                    # Also store the relative path for user convenience.
                    # Don't use pathlib.Path.relative_to, because that
                    # only has `walk_up` since Python 3.12.
                    # Also, just look in subdirectories which simplifies
                    # path resolution.
                    this_parent = str(self.path.parent) + os.sep
                    path_parent = str(pp.parent) + os.sep
                    if path_parent.startswith(this_parent):
                        flocs.append(str(pp).replace(this_parent, "", 1))
                else:
                    # We already did (or did not upon user request) verify
                    # the path. Just pass it on to the list.
                    flocs.append(str(pp))
            b_data["paths"] = flocs
        elif basin_type == "internal":
            b_data["paths"] = basin_locs
        elif basin_type == "remote":
            b_data["urls"] = [str(p) for p in basin_locs]
        else:
            raise ValueError(f"Unknown basin type '{basin_type}'")

        b_lines = json.dumps(b_data, indent=2, sort_keys=True).split("\n")
        basins = self.h5file.require_group("basins")
        key = hashobj(b_lines)
        if key not in basins:
            self.write_text(basins, key, b_lines)
        return key

    def store_feature(self, feat, data, shape=None):
        """Write feature data

        Parameters
        ----------
        feat: str
            feature name
        data: np.ndarray or list or dict
            feature data
        shape: tuple of int
            For non-scalar features, this is the shape of the
            feature for one event (e.g. `(90, 250)` for an "image".
            Usually, you do not have to specify this value, but you
            do need it in case of plugin features that don't have
            the "feature shape" set or in case of temporary features.
            If you don't specify it, then the shape is guessed based
            on the data you provide and a UserWarning will be issued.
        """
        if not dfn.feature_exists(feat):
            raise ValueError(f"Undefined feature '{feat}'!")

        events = self.h5file.require_group("events")

        # replace data?
        if feat in events and self.mode == "replace":
            if feat == "trace":
                for tr_name in data.keys():
                    if tr_name in events[feat]:
                        del events[feat][tr_name]
            else:
                del events[feat]

        if feat in FEATURES_UINT32:
            dtype = np.uint32
        elif feat in FEATURES_UINT64:
            dtype = np.uint64
        else:
            dtype = None

        if feat == "index":
            # By design, the index must be a simple enumeration.
            # We enforce that by not trusting the user. If you need
            # a different index, please take a look at the index_online
            # feature.
            nev = len(data)
            if "index" in events:
                nev0 = len(events["index"])
            else:
                nev0 = 0
            self.write_ndarray(group=events,
                               name="index",
                               data=np.arange(nev0 + 1, nev0 + nev + 1),
                               dtype=dtype)
        elif dfn.scalar_feature_exists(feat):
            self.write_ndarray(group=events,
                               name=feat,
                               data=np.atleast_1d(data),
                               dtype=dtype)
        elif feat == "contour":
            self.write_ragged(group=events, name=feat, data=data)
        elif feat in ["image", "image_bg", "mask", "qpi_oah", "qpi_oah_bg"]:
            self.write_image_grayscale(group=events,
                                       name=feat,
                                       data=data,
                                       is_boolean=(feat == "mask"))
        elif feat in ["qpi_amp", "qpi_pha"]:
            self.write_image_float32(group=events,
                                     name=feat,
                                     data=data)
        elif feat == "trace":
            for tr_name in data.keys():
                # verify trace names
                if tr_name not in dfn.FLUOR_TRACES:
                    raise ValueError(f"Unknown trace key: '{tr_name}'!")
                # write trace
                self.write_ndarray(group=events.require_group("trace"),
                                   name=tr_name,
                                   data=np.atleast_2d(data[tr_name]),
                                   dtype=dtype
                                   )
        else:
            if not shape:
                # OK, so we are dealing with a plugin feature or a temporary
                # feature here. Now, we don't know the exact shape of that
                # feature, but we give the user the option to advertise
                # the shape of the feature in the plugin.
                # First, try to obtain the shape from the PluginFeature
                # (if that exists).
                for pf in PlugInFeature.get_instances(feat):
                    if isinstance(pf, PlugInFeature):
                        shape = pf.plugin_feature_info.get("feature shape")
                        if shape is not None:
                            break  # This is good.
                else:
                    # Temporary features will have to live with this warning.
                    warnings.warn(
                        "There is no information about the shape of the "
                        + f"feature '{feat}'. I am going out on a limb "
                        + "for you and assume that you are storing "
                        + "multiple events at a time. If this works, "
                        + f"you could put the shape `{data[0].shape}` "
                        + 'in the `info["feature shapes"]` key of '
                        + "your plugin feature.")
                    shape = data.shape[1:]
            if shape == data.shape:
                data = data.reshape(1, *shape)
            elif shape == data.shape[1:]:
                pass
            else:
                raise ValueError(f"Bad shape for {feat}! Expeted {shape}, "
                                 + f"but got {data.shape[1:]}!")
            self.write_ndarray(group=events, name=feat, data=data, dtype=dtype)

    def store_log(self, name, lines):
        """Write log data

        Parameters
        ----------
        name: str
            name of the log entry
        lines: list of str or str
            the text lines of the log
        """
        log_group = self.h5file.require_group("logs")
        self.write_text(group=log_group, name=name, lines=lines)

    def store_metadata(self, meta):
        """Store RT-DC metadata

        Parameters
        ----------
        meta: dict-like
            The metadata to store. Each key depicts a metadata section
            name whose data is given as a dictionary, e.g.::

                meta = {"imaging": {"exposure time": 20,
                                    "flash duration": 2,
                                    ...
                                    },
                        "setup": {"channel width": 20,
                                  "chip region": "channel",
                                  ...
                                  },
                        ...
                        }

            Only section key names and key values therein registered
            in dclab are allowed and are converted to the pre-defined
            dtype. Only sections from the
            :const:`dclab.definitions.CFG_METADATA` dictionary are
            stored. If you have custom metadata, you can use the "user"
            section.
        """
        meta = copy.deepcopy(meta)
        # Ignore/remove tdms section
        meta.pop("fmt_tdms", None)
        # Check meta data
        for sec in meta:
            if sec == "user":
                # user-defined metadata are always written.
                # Any errors (incompatibilities with HDF5 attributes)
                # are the user's responsibility
                continue
            elif sec not in dfn.CFG_METADATA:
                # only allow writing of meta data that are not editable
                # by the user (not dclab.dfn.CFG_ANALYSIS)
                raise ValueError(
                    f"Meta data section not defined in dclab: {sec}")
            for ck in meta[sec]:
                if not dfn.config_key_exists(sec, ck):
                    raise ValueError(
                        f"Meta key not defined in dclab: {sec}:{ck}")

        # update version
        old_version = meta.get("setup", {}).get("software version", "")
        new_version = self.version_brand(
            old_version=old_version or None,
            write_attribute=False
        )
        meta.setdefault("setup", {})["software version"] = new_version

        # Write metadata
        for sec in meta:
            for ck in meta[sec]:
                idk = f"{sec}:{ck}"
                value = meta[sec][ck]
                if isinstance(value, bytes):
                    # We never store byte attribute values.
                    # In this case, `convfunc` should be `str` or `lcstr` or
                    # somesuch. But we don't test that, because no other
                    # datatype competes with str for bytes.
                    value = value.decode("utf-8")
                if sec == "user":
                    # store user-defined metadata as-is
                    self.h5file.attrs[idk] = value
                else:
                    # pipe the metadata through the hard-coded converter
                    # functions
                    convfunc = dfn.get_config_value_func(sec, ck)
                    self.h5file.attrs[idk] = convfunc(value)

    def store_table(self, name, cmp_array, h5_attrs=None):
        """Store a compound array table

        Tables are semi-metadata. They may contain information collected
        during a measurement (but with a lower temporal resolution) or
        other tabular data relevant for a dataset. Tables have named
        columns. Therefore, they can be represented as a numy recarray,
        and they should be stored as such in an HDF5 file (compund dataset).

        Parameters
        ----------
        name: str
            Name of the table
        cmp_array: np.recarray, h5py.Dataset, np.ndarray, or dict
            If a np.recarray or h5py.Dataset are provided, then they
            are written as-is to the file. If a dictionary is provided,
            then the dictionary is converted into a numpy recarray.
            If a numpy array is provided, then the array is written
            as a raw table (no column names) to the file.
        h5_attrs: dict, optional
            Attributes to store alongside the corresponding HDF5 dataset
        """
        if h5_attrs is None:
            h5_attrs = {}

        # Convert MetaTable to numpy data
        if isinstance(cmp_array, MetaTable):
            h5_attrs.update(cmp_array.meta)
            cmp_array = cmp_array.__array__()

        # Handle individual cases
        if isinstance(cmp_array, np.recarray):
            # A table is a compound array (np.recarray). If we are here,
            # this means that the user passed an instance of np.recarray.
            pass
        elif isinstance(cmp_array, h5py.Dataset):
            # An instance of h5py.Dataset (which we trust to be a proper
            # compound dataset at this point). No additional steps needed.
            h5_attrs.update(cmp_array.attrs)
            pass
        elif isinstance(cmp_array, np.ndarray):
            # A numpy array was passed. This usually means we have something
            # that we can look at, so we add image tags.
            h5_attrs['CLASS'] = np.bytes_('IMAGE')
            h5_attrs['IMAGE_VERSION'] = np.bytes_('1.2')
            h5_attrs['IMAGE_SUBCLASS'] = np.bytes_('IMAGE_GRAYSCALE')
            pass
        elif isinstance(cmp_array, dict):
            # The user passed a dict which we now have to convert to a
            # compound dataset. We do this because we are user-convenient.
            # The user should not need to wade through these steps:
            columns = list(cmp_array.keys())
            # Everything should be floats in a table.
            ds_dt = np.dtype({'names': columns,
                              'formats': [np.float64] * len(columns)})
            # We trust the user to provide a dictionary with one-dimensional
            # lists or arrays of the same length.
            tabsize = len(cmp_array[columns[0]])
            tab_data = np.zeros((tabsize, len(columns)))
            for ii, tab in enumerate(columns):
                tab_data[:, ii] = cmp_array[tab]
            # Now create a new compound array (discarding the old dict)
            cmp_array = np.rec.array(tab_data, dtype=ds_dt)
        else:
            raise NotImplementedError(
                f"Cannot convert {type(cmp_array)} to table!")

        # data
        group = self.h5file.require_group("tables")
        tab = group.create_dataset(
            name,
            data=cmp_array,
            fletcher32=True,
            **self.compression_kwargs)

        # metadata
        if h5_attrs:
            tab.attrs.update(h5_attrs)

    def version_brand(self, old_version=None, write_attribute=True):
        """Perform version branding

        Append a " | dclab X.Y.Z" to the "setup:software version"
        attribute.

        Parameters
        ----------
        old_version: str or None
            By default, the version string is taken from the HDF5 file.
            If set to a string, then this version is used instead.
        write_attribute: bool
            If True (default), write the version string to the
            "setup:software version" attribute
        """
        if old_version is None:
            old_version = self.h5file.attrs.get("setup:software version", "")
        if isinstance(old_version, bytes):
            old_version = old_version.decode("utf-8")
        version_chain = [vv.strip() for vv in old_version.split("|")]
        version_chain = [vv for vv in version_chain if vv]
        cur_version = "dclab {}".format(version)

        if version_chain:
            if version_chain[-1] != cur_version:
                version_chain.append(cur_version)
        else:
            version_chain = [cur_version]
        new_version = " | ".join(version_chain)
        if write_attribute:
            self.h5file.attrs["setup:software version"] = new_version
        else:
            return new_version

    def write_image_float32(self, group, name, data):
        """Write 32bit floating point image array

        This function wraps :func:`RTDCWriter.write_ndarray`
        and adds image attributes to the HDF5 file so HDFView
        can display the images properly.

        Parameters
        ----------
        group: h5py.Group
            parent group
        name: str
            name of the dataset containing the text
        data: np.ndarray or list of np.ndarray
            image data
        """
        if isinstance(data, (list, tuple)):
            # images may be in lists
            data = np.atleast_2d(data)

        if len(data.shape) == 2:
            # put single event in 3D array
            data = data[np.newaxis]

        dset = self.write_ndarray(group=group, name=name, data=data,
                                  dtype=np.float32)

        # Create and Set image attributes:
        # HDFView recognizes this as a series of images.
        # Use np.bytes_ as per
        # https://docs.h5py.org/en/stable/strings.html#compatibility
        dset.attrs.create('CLASS', np.bytes_('IMAGE'))
        dset.attrs.create('IMAGE_VERSION', np.bytes_('1.2'))
        dset.attrs.create('IMAGE_SUBCLASS', np.bytes_('IMAGE_GRAYSCALE'))

    def write_image_grayscale(self, group, name, data, is_boolean):
        """Write grayscale image data to and HDF5 dataset

        This function wraps :func:`RTDCWriter.write_ndarray`
        and adds image attributes to the HDF5 file so HDFView
        can display the images properly.

        Parameters
        ----------
        group: h5py.Group
            parent group
        name: str
            name of the dataset containing the text
        data: np.ndarray or list of np.ndarray
            image data
        is_boolean: bool
            whether the input data is of boolean nature
            (e.g. mask data) - if so, data are converted to uint8
        """
        if isinstance(data, (list, tuple)):
            # images may be in lists
            data = np.atleast_2d(data)

        if len(data.shape) == 2:
            # put single event in 3D array
            data = data.reshape(1, data.shape[0], data.shape[1])

        if is_boolean:
            # convert binary (mask) data to uint8
            if data.__class__.__name__ == "H5MaskEvent":
                # (if we use `isinstance`, we get circular imports)
                # Be smart and directly write back the original data
                # (otherwise we would convert to bool and back to uint8).
                data = data.h5dataset
            elif data.dtype == bool:
                # Convert binary input mask data to uint8 with max range
                data = np.asarray(data, dtype=np.uint8) * 255

        dset = self.write_ndarray(group=group, name=name, data=data,
                                  dtype=np.uint8)

        # Create and Set image attributes:
        # HDFView recognizes this as a series of images.
        # Use np.bytes_ as per
        # https://docs.h5py.org/en/stable/strings.html#compatibility
        dset.attrs.create('CLASS', np.bytes_('IMAGE'))
        dset.attrs.create('IMAGE_VERSION', np.bytes_('1.2'))
        dset.attrs.create('IMAGE_SUBCLASS', np.bytes_('IMAGE_GRAYSCALE'))

    def write_ndarray(self, group, name, data, dtype=None):
        """Write n-dimensional array data to an HDF5 dataset

        It is assumed that the shape of the array data is correct,
        i.e. that the shape of `data` is
        (number_events, feat_shape_1, ..., feat_shape_n).

        Parameters
        ----------
        group: h5py.Group
            parent group
        name: str
            name of the dataset containing the text
        data: np.ndarray
            data
        dtype: dtype
            the dtype to use for storing the data
            (defaults to `data.dtype`)
        """
        if len(data) == 0:
            raise ValueError(f"Empty data object for '{name}'")

        if name not in group:
            chunks = self.get_best_nd_chunks(item_shape=data.shape[1:],
                                             item_dtype=data.dtype)
            maxshape = tuple([None] + list(data.shape)[1:])
            dset = group.create_dataset(
                name,
                shape=data.shape,
                dtype=dtype or data.dtype,
                maxshape=maxshape,
                chunks=chunks,
                fletcher32=True,
                **self.compression_kwargs)
            offset = 0
        else:
            dset = group[name]
            offset = dset.shape[0]
            dset.resize(offset + data.shape[0], axis=0)
        if len(data.shape) == 1:
            # store scalar data in one go
            dset[offset:] = data
            # store ufunc data for min/max
            for uname, ufunc in [("min", np.nanmin),
                                 ("max", np.nanmax)]:
                val_a = dset.attrs.get(uname, None)
                if val_a is not None:
                    val_b = ufunc(data)
                    val = ufunc([val_a, val_b])
                else:
                    val = ufunc(dset)
                dset.attrs[uname] = val
            # store ufunc data for mean (weighted with size)
            mean_a = dset.attrs.get("mean", None)
            if mean_a is not None:
                num_a = offset
                mean_b = np.nanmean(data)
                num_b = data.size
                mean = (mean_a * num_a + mean_b * num_b) / (num_a + num_b)
            else:
                mean = np.nanmean(dset)
            dset.attrs["mean"] = mean
        else:
            chunk_size = dset.chunks[0]
            # populate higher-dimensional data in chunks
            # (reduces file size, memory usage, and saves time)
            num_chunks = len(data) // chunk_size
            for ii in range(num_chunks):
                start = ii * chunk_size
                stop = start + chunk_size
                dset[offset+start:offset+stop] = data[start:stop]
            # write remainder (if applicable)
            num_remain = len(data) % chunk_size
            if num_remain:
                start_e = num_chunks * chunk_size
                stop_e = start_e + num_remain
                dset[offset+start_e:offset+stop_e] = data[start_e:stop_e]
        return dset

    def write_ragged(self, group, name, data):
        """Write ragged data (i.e. list of arrays of different lenghts)

        Ragged array data (e.g. contour data) are stored in
        a separate group and each entry becomes an HDF5 dataset.

        Parameters
        ----------
        group: h5py.Group
            parent group
        name: str
            name of the dataset containing the text
        data: list of np.ndarray or np.ndarray
            the data in a list
        """
        if isinstance(data, np.ndarray) and len(data.shape) == 2:
            # place single event in list
            data = [data]
        grp = group.require_group(name)
        # The following case is just a workaround for the very slow
        # `len(grp)` which makes things horrible if you are storing
        # contour data one-by-one. The only downside of this is that
        # we have to keep track of the length of the group. But I
        # think that is OK, since everything is very private here.
        # - Paul (2021-10-18)
        if grp not in self._group_sizes:
            self._group_sizes[grp] = len(grp)
        curid = self._group_sizes[grp]
        for ii, cc in enumerate(data):
            grp.create_dataset("{}".format(curid + ii),
                               data=cc,
                               fletcher32=True,
                               chunks=cc.shape,
                               **self.compression_kwargs)
            self._group_sizes[grp] += 1

    def write_text(self, group, name, lines):
        """Write text to an HDF5 dataset

        Text data are written as a fixed-length string dataset.

        Parameters
        ----------
        group: h5py.Group
            parent group
        name: str
            name of the dataset containing the text
        lines: list of str or str
            the text, line by line
        """
        # replace text?
        if name in group and self.mode == "replace":
            del group[name]

        # handle strings
        if isinstance(lines, (str, bytes)):
            lines = [lines]

        lnum = len(lines)
        # Determine the maximum line length and use fixed-length strings,
        # because compression and fletcher32 filters won't work with
        # variable length strings.
        # https://github.com/h5py/h5py/issues/1948
        # 100 is the recommended maximum and the default, because if
        # `mode` is e.g. "append", then this line may not be the longest.
        max_length = 100
        lines_as_bytes = []
        for line in lines:
            # convert lines to bytes
            if not isinstance(line, bytes):
                lbytes = line.encode("UTF-8")
            else:
                lbytes = line
            max_length = max(max_length, len(lbytes))
            lines_as_bytes.append(lbytes)

        if name not in group:
            # Create the dataset
            txt_dset = group.create_dataset(
                name,
                shape=(lnum,),
                dtype=f"S{max_length}",
                maxshape=(None,),
                chunks=True,
                fletcher32=True,
                **self.compression_kwargs)
            line_offset = 0
        else:
            # TODO: test whether fixed length is long enough!
            # Resize the dataset
            txt_dset = group[name]
            line_offset = txt_dset.shape[0]
            txt_dset.resize(line_offset + lnum, axis=0)

        # Write the text data line-by-line
        for ii, lbytes in enumerate(lines_as_bytes):
            txt_dset[line_offset + ii] = lbytes



"""Kernel Density Estimation methods"""

import numpy as np
from scipy.interpolate import RectBivariateSpline
from scipy.stats import gaussian_kde, skew

from .cached import Cache
from .external.statsmodels.nonparametric.kernel_density import KDEMultivariate


def bin_num_doane(a):
    """Compute number of bins based on Doane's formula

    Notes
    -----
    If the bin width cannot be determined, then a bin
    number of 5 is returned.

    See Also
    --------
    bin_width_doane: method used to compute the bin width
    """
    bad = np.isnan(a) | np.isinf(a)
    data = a[~bad]
    acc = bin_width_doane(a)
    if acc == 0 or np.isnan(acc):
        num = 5
    else:
        num = int(np.round((data.max() - data.min()) / acc))
    return num


def bin_width_doane(a):
    """Compute contour spacing based on Doane's formula

    References
    ----------
    - `<https://en.wikipedia.org/wiki/Histogram#Number_of_bins_and_width>`_
    - `<https://stats.stackexchange.com/questions/55134/
      doanes-formula-for-histogram-binning>`_

    Notes
    -----
    Doane's formula is actually designed for histograms. This
    function is kept here for backwards-compatibility reasons.
    It is highly recommended to use :func:`bin_width_percentile`
    instead.
    """
    bad = np.isnan(a) | np.isinf(a)
    data = a[~bad]
    n = data.size
    g1 = skew(data)
    sigma_g1 = np.sqrt(6 * (n - 2) / ((n + 1) * (n + 3)))
    k = 1 + np.log2(n) + np.log2(1 + np.abs(g1) / sigma_g1)
    acc = (data.max() - data.min()) / k
    return acc


def bin_width_percentile(a):
    """Compute contour spacing based on data percentiles

    The 10th and the 90th percentile of the input data are taken.
    The spacing then computes to the difference between those
    two percentiles divided by 23.

    Notes
    -----
    The Freedman–Diaconis rule uses the interquartile range and
    normalizes to the third root of len(a). Such things do not
    work very well for RT-DC data, because len(a) is huge. Here
    we use just the top and bottom 10th percentiles with a fixed
    normalization.
    """
    bad = np.isnan(a) | np.isinf(a)
    data = a[~bad]
    start = np.percentile(data, 10)
    end = np.percentile(data, 90)
    acc = (end - start) / 23
    return acc


def get_bad_vals(x, y):
    return np.isnan(x) | np.isinf(x) | np.isnan(y) | np.isinf(y)


def ignore_nan_inf(kde_method):
    """Ignores nans and infs from the input data

    Invalid positions in the resulting density are set to nan.
    """
    def new_kde_method(events_x, events_y, xout=None, yout=None,
                       *args, **kwargs):
        bad_in = get_bad_vals(events_x, events_y)
        if xout is None:
            density = np.zeros_like(events_x, dtype=np.float64)
            bad_out = bad_in
            xo = yo = None
        else:
            density = np.zeros_like(xout, dtype=np.float64)
            bad_out = get_bad_vals(xout, yout)
            xo = xout[~bad_out]
            yo = yout[~bad_out]
        # Filter events
        ev_x = events_x[~bad_in]
        ev_y = events_y[~bad_in]
        density[~bad_out] = kde_method(ev_x, ev_y,
                                       xo, yo,
                                       *args, **kwargs)
        density[bad_out] = np.nan
        return density

    doc_add = "\n    Notes\n" +\
              "    -----\n" +\
              "    This is a wrapped version that ignores nan and inf values."
    new_kde_method.__doc__ = kde_method.__doc__ + doc_add

    return new_kde_method


@ignore_nan_inf
@Cache
def kde_gauss(events_x, events_y, xout=None, yout=None):
    """ Gaussian Kernel Density Estimation

    Parameters
    ----------
    events_x, events_y: 1D ndarray
        The input points for kernel density estimation. Input
        is flattened automatically.
    xout, yout: ndarray
        The coordinates at which the KDE should be computed.
        If set to none, input coordinates are used.

    Returns
    -------
    density: ndarray, same shape as `xout`
        The KDE for the points in (xout, yout)

    See Also
    --------
    `scipy.stats.gaussian_kde`
    """
    valid_combi = ((xout is None and yout is None) or
                   (xout is not None and yout is not None)
                   )
    if not valid_combi:
        raise ValueError("Both `xout` and `yout` must be (un)set.")

    if xout is None and yout is None:
        xout = events_x
        yout = events_y

    try:
        estimator = gaussian_kde([events_x.flatten(), events_y.flatten()])
        density = estimator.evaluate([xout.flatten(), yout.flatten()])
    except np.linalg.LinAlgError:
        # LinAlgError occurs when matrix to solve is singular (issue #117)
        density = np.zeros(xout.shape)*np.nan
    return density.reshape(xout.shape)


@ignore_nan_inf
@Cache
def kde_histogram(events_x, events_y, xout=None, yout=None, bins=None):
    """ Histogram-based Kernel Density Estimation

    Parameters
    ----------
    events_x, events_y: 1D ndarray
        The input points for kernel density estimation. Input
        is flattened automatically.
    xout, yout: ndarray
        The coordinates at which the KDE should be computed.
        If set to none, input coordinates are used.
    bins: tuple (binsx, binsy)
        The number of bins to use for the histogram.

    Returns
    -------
    density: ndarray, same shape as `xout`
        The KDE for the points in (xout, yout)

    See Also
    --------
    `numpy.histogram2d`
    `scipy.interpolate.RectBivariateSpline`
    """
    valid_combi = ((xout is None and yout is None) or
                   (xout is not None and yout is not None)
                   )
    if not valid_combi:
        raise ValueError("Both `xout` and `yout` must be (un)set.")

    if xout is None and yout is None:
        xout = events_x
        yout = events_y

    if bins is None:
        bins = (max(5, bin_num_doane(events_x)),
                max(5, bin_num_doane(events_y)))

    # Compute the histogram
    hist2d, xedges, yedges = np.histogram2d(x=events_x,
                                            y=events_y,
                                            bins=bins,
                                            density=True)
    xip = xedges[1:]-(xedges[1]-xedges[0])/2
    yip = yedges[1:]-(yedges[1]-yedges[0])/2

    estimator = RectBivariateSpline(x=xip, y=yip, z=hist2d)
    density = estimator.ev(xout, yout)
    density[density < 0] = 0

    return density.reshape(xout.shape)


def kde_none(events_x, events_y, xout=None, yout=None):
    """No Kernel Density Estimation

    Parameters
    ----------
    events_x, events_y: 1D ndarray
        The input points for kernel density estimation. Input
        is flattened automatically.
    xout, yout: ndarray
        The coordinates at which the KDE should be computed.
        If set to none, input coordinates are used.

    Returns
    -------
    density: ndarray, same shape as `xout`
        The KDE for the points in (xout, yout)

    Notes
    -----
    This method is a convenience method that always returns ones in the shape
    that the other methods in this module produce.
    """
    valid_combi = ((xout is None and yout is None) or
                   (xout is not None and yout is not None)
                   )
    if not valid_combi:
        raise ValueError("Both `xout` and `yout` must be (un)set.")

    if xout is None and yout is None:
        xout = events_x
        _ = events_y

    return np.ones(xout.shape)


@ignore_nan_inf
@Cache
def kde_multivariate(events_x, events_y, xout=None, yout=None, bw=None):
    """ Multivariate Kernel Density Estimation

    Parameters
    ----------
    events_x, events_y: 1D ndarray
        The input points for kernel density estimation. Input
        is flattened automatically.
    bw: tuple (bwx, bwy) or None
        The bandwith for kernel density estimation.
    xout, yout: ndarray
        The coordinates at which the KDE should be computed.
        If set to none, input coordinates are used.

    Returns
    -------
    density: ndarray, same shape as `xout`
        The KDE for the points in (xout, yout)

    See Also
    --------
    `statsmodels.nonparametric.kernel_density.KDEMultivariate`
    """
    valid_combi = ((xout is None and yout is None) or
                   (xout is not None and yout is not None)
                   )
    if not valid_combi:
        raise ValueError("Both `xout` and `yout` must be (un)set.")

    if xout is None and yout is None:
        xout = events_x
        yout = events_y
    if bw is None:
        # divide by 2 to make it comparable to histogram KDE
        bw = (bin_width_doane(events_x) / 2,
              bin_width_doane(events_y) / 2)

    positions = np.vstack([xout.flatten(), yout.flatten()])
    estimator_ly = KDEMultivariate(data=[events_x.flatten(),
                                         events_y.flatten()],
                                   var_type='cc', bw=bw)

    density = estimator_ly.pdf(positions)
    return density.reshape(xout.shape)


methods = {"gauss": kde_gauss,
           "histogram": kde_histogram,
           "none": kde_none,
           "multivariate": kde_multivariate}



"""RT-DC dictionary format"""
import collections
import time

import numpy as np

from .. import definitions as dfn
from ..util import hashobj

from .config import Configuration
from .core import RTDCBase


class DictContourEvent:
    def __init__(self, contours):
        assert contours[0].shape[1] == 2
        self.shape = (len(contours), np.nan, 2)
        self.contours = contours

    def __iter__(self):
        return iter(self.contours)

    def __getitem__(self, item):
        return self.contours[item]

    def __len__(self):
        return len(self.contours)


class DictTraceEvent(collections.UserDict):

    @property
    def shape(self):
        key0 = sorted(self.keys())[0]
        return len(self), len(self[key0]), len(self[key0][0])


class RTDC_Dict(RTDCBase):
    _local_basins_allowed = True

    def __init__(self, ddict, *args, **kwargs):
        """Dictionary-based RT-DC dataset

        Parameters
        ----------
        ddict: dict
            Dictionary with features as keys (valid features like
            "area_cvx", "deform", "image" are defined by
            `dclab.definitions.feature_exists`) with which the class
            will be instantiated. The configuration is set to the
            default configuration of dclab.

            .. versionchanged:: 0.27.0
                Scalar features are automatically converted to arrays.
        *args:
            Arguments for `RTDCBase`
        **kwargs:
            Keyword arguments for `RTDCBase`
        """
        assert ddict

        super(RTDC_Dict, self).__init__(*args, **kwargs)

        t = time.localtime()

        # Get an identifying string
        keys = list(ddict.keys())
        keys.sort()
        ids = hashobj(ddict[keys[0]])
        self._ids = ids
        self.path = "none"
        self.title = "{}_{:02d}_{:02d}/{}.dict".format(t[0], t[1], t[2], ids)

        # Populate events
        for feat in ddict:
            if dfn.feature_exists(feat):
                if dfn.scalar_feature_exists(feat):
                    data = np.array(ddict[feat])
                elif feat == "contour":
                    data = DictContourEvent(ddict[feat])
                elif feat == "trace":
                    data = DictTraceEvent(ddict[feat])
                elif isinstance(ddict[feat], list):
                    # convert e.g. image data to arrays
                    data = np.array(ddict[feat])
                else:
                    data = ddict[feat]
            else:
                raise ValueError("Invalid feature name '{}'".format(feat))
            if isinstance(data, np.ndarray):
                # Convert numpy array to read-only array
                data = data.view()
                data.setflags(write=False)
            self._events[feat] = data

        event_count = len(ddict[list(ddict.keys())[0]])

        self.config = Configuration()
        self.config["experiment"]["event count"] = event_count

    @property
    def hash(self):
        return self._ids



"""Computation of apparent Young's modulus for RT-DC measurements"""
from __future__ import annotations

import numbers
import pathlib
from typing import Literal
import warnings

import numpy as np
import scipy.interpolate as spint

from ...warn import PipelineWarning
from .load import load_lut, load_mtext, register_lut  # noqa: F401
from .pxcorr import get_pixelation_delta
from .pxcorr import get_pixelation_delta_pair  # noqa: F401
# TODO: remove deprecated `convert`
from .scale_linear import convert  # noqa: F401
from .scale_linear import scale_emodulus, scale_feature
from .viscosity import get_viscosity


#: Set this to True to globally enable spline extrapolation when the
#: `area_um`/`deform` data are outside the LUT. This is discouraged and
#: a :class:`KnowWhatYouAreDoingWarning` warning will be issued.
INACCURATE_SPLINE_EXTRAPOLATION = False


class KnowWhatYouAreDoingWarning(PipelineWarning):
    pass


class YoungsModulusLookupTableExceededWarning(PipelineWarning):
    pass


def extrapolate_emodulus(lut, datax, deform, emod, deform_norm,
                         deform_thresh=.05, inplace=True):
    """Use spline interpolation to fill in nan-values

    When points (`datax`, `deform`) are outside the convex
    hull of the lut, then :func:`scipy.interpolate.griddata` returns
    nan-valules.

    With this function, some of these nan-values are extrapolated
    using :class:`scipy.interpolate.SmoothBivariateSpline`. The
    supported extrapolation values are currently limited to those
    where the deformation is above 0.05.

    A warning will be issued, because this is not really
    recommended.

    Parameters
    ----------
    lut: ndarray of shape (N, 3)
        The normalized (!! see :func:`normalize`) LUT (first axis is
        points, second axis enumerates datax, deform, and emodulus)
    datax: ndarray of size N
        The normalized x data (corresponding to `lut[:, 0]`)
    deform: ndarray of size N
        The normalized deform (corresponding to `lut[:, 1]`)
    emod: ndarray of size N
        The emodulus (corresponding to `lut[:, 2]`); If `emod`
        does not contain nan-values, there is nothing to do here.
    deform_norm: float
        The normalization value used to normalize `lut[:, 1]` and
        `deform`.
    deform_thresh: float
        Not the entire LUT is used for bivariate spline interpolation.
        Only the points where `lut[:, 1] > deform_thresh/deform_norm`
        are used. This is necessary, because for small deformations,
        the LUT has an extreme slope that kills any meaningful
        spline interpolation.
    inplace: bool
        If True (default), replaces nan values in `emod` in-place.
        If False, `emod` is not modified.
    """
    if not inplace:
        emod = np.array(emod, copy=True)
    # unknowns are nans and deformation values above the threshold
    unkn = np.logical_and(np.isnan(emod),
                          deform > deform_thresh/deform_norm)

    if np.sum(unkn) == 0:
        # nothing to do
        return emod

    warnings.warn("LUT extrapolation is barely tested and may yield "
                  + "unphysical values!",
                  KnowWhatYouAreDoingWarning)

    lut_crop = lut[lut[:, 1] > deform_thresh/deform_norm, :]

    itp = spint.SmoothBivariateSpline(x=lut_crop[:, 0],
                                      y=lut_crop[:, 1],
                                      z=lut_crop[:, 2],
                                      )

    emod[unkn] = itp.ev(datax[unkn], deform[unkn])
    return emod


def get_emodulus(deform: float | np.array,
                 area_um: float | np.array | None = None,
                 volume: float | np.array | None = None,
                 medium: float | str = "0.49% MC-PBS",
                 channel_width: float = 20.0,
                 flow_rate: float = 0.16,
                 px_um: float = 0.34,
                 temperature: float | np.ndarray | None = 23.0,
                 lut_data: str | pathlib.Path | np.ndarray = "LE-2D-FEM-19",
                 visc_model: Literal['herold-2017',
                                     'herold-2017-fallback',
                                     'buyukurganci-2022',
                                     'kestin-1978',
                                     None] = "herold-2017-fallback",
                 extrapolate: bool = INACCURATE_SPLINE_EXTRAPOLATION,
                 copy: bool = True):
    """Compute apparent Young's modulus using a look-up table

    Parameters
    ----------
    area_um: float or ndarray
        Apparent (2D image) area [µm²] of the event(s)
    deform: float or ndarray
        Deformation (1-circularity) of the event(s)
    volume: float or ndarray
        Apparent volume of the event(s). It is not possible to define
        `volume` and `area_um` at the same time (makes no sense).

        .. versionadded:: 0.25.0
    medium: str or float
        The medium to compute the viscosity for. If a string
        is given, the viscosity is computed. If a float is given,
        this value is used as the viscosity in mPa*s (Note that
        `temperature` and `visc_model` must be set to None in this case).
    channel_width: float
        The channel width [µm]
    flow_rate: float
        Flow rate [µL/s]
    px_um: float
        The detector pixel size [µm] used for pixelation correction.
        Set to zero to disable.
    temperature: float, ndarray, or None
        Temperature [°C] of the event(s)
    lut_data: path, str, or tuple of (np.ndarray of shape (N, 3), dict)
        The LUT data to use. If it is a built-in identifier,
        then the respective LUT will be used. Otherwise, a path to a
        file on disk or a tuple (LUT array, metadata) is possible.
        The LUT metadata is used to check whether the given features
        (e.g. `area_um` and `deform`) are valid interpolation choices.

        .. versionadded:: 0.25.0
    visc_model: str
        The viscosity model to use,
        see :func:`dclab.features.emodulus.viscosity.get_viscosity`
    extrapolate: bool
        Perform extrapolation using :func:`extrapolate_emodulus`. This
        is discouraged!
    copy: bool
        Copy input arrays. If set to false, input arrays are
        overridden.

    Returns
    -------
    elasticity: float or ndarray
        Apparent Young's modulus in kPa

    Notes
    -----
    - The look-up table used was computed with finite elements methods
      according to :cite:`Mokbel2017` and complemented with analytical
      isoelastics from :cite:`Mietke2015`. The original simulation
      results are available on figshare :cite:`FigshareWittwer2020`.
    - The computation of the Young's modulus takes into account
      a correction for the viscosity (medium, channel width, flow rate,
      and temperature) :cite:`Mietke2015` and a correction for
      pixelation for the deformation which were derived
      from a (pixelated) image :cite:`Herold2017`.
    - Note that while deformation is pixelation-corrected, area_um and
      volume are scaled to match the LUT data. This is somewhat
      fortunate, because we don't have to worry about the order of
      applying pixelation correction and scale conversion.
    - By using external LUTs, it is possible to interpolate on the
      volume-deformation plane. This feature was added in version
      0.25.0.

    See Also
    --------
    dclab.features.emodulus.viscosity.get_viscosity: compute viscosity
        for known media
    """
    # Get lut data
    lut, lut_meta = load_lut(lut_data)
    # Get the correct features (sanity checks)
    featx, featy, _ = lut_meta["column features"]
    if featx == "area_um" and featy == "deform":
        assert volume is None, "Don't define area_um and volume at same time!"
        datax = np.array(area_um, dtype=float, copy=copy)
    elif featx == "volume" and featy == "deform":
        assert area_um is None, "Don't define area_um and volume at same time!"
        datax = np.array(volume, dtype=float, copy=copy)
    else:
        raise KeyError("No recipe for '{}' and '{}'!".format(featx, featy))

    # copy deform so we can use in-place pixel correction
    deform = np.array(deform, dtype=float, copy=copy)
    if px_um:
        # Correct deformation for pixelation effect (subtract ddelt).
        # It is important to do this before scaling datax (area_um/volume).
        deform -= get_pixelation_delta(feat_corr=featy,
                                       feat_absc=featx,
                                       data_absc=datax,
                                       px_um=px_um)

    # Compute viscosity
    if isinstance(medium, numbers.Number):
        visco = medium
        if temperature is not None:
            raise ValueError("If `medium` is given in Pa*s, then "
                             + "`temperature` must be set to None!")
        if visc_model is not None:
            warnings.warn("If `medium` is given in Pa*s, then `visc_model` "
                          "must be set to None. An exception will be raised "
                          "in future versions of dclab.",
                          DeprecationWarning)
    else:
        visco = get_viscosity(medium=medium, channel_width=channel_width,
                              flow_rate=flow_rate, temperature=temperature,
                              model=visc_model)

    if isinstance(visco, np.ndarray):
        # New in dclab 0.20.0: Computation for viscosities array
        # Convert the input area_um to that of the LUT (deform does not change)
        scale_kw = {"channel_width_in": channel_width,
                    "channel_width_out": lut_meta["channel_width"],
                    "inplace": False}
        datax_4lut = scale_feature(feat=featx, data=datax, **scale_kw)
        deform_4lut = np.array(deform, dtype=float, copy=copy)

        # Normalize interpolation data such that the spacing for
        # area and deformation is about the same during interpolation.
        featx_norm = lut[:, 0].max()
        normalize(lut[:, 0], featx_norm)
        normalize(datax_4lut, featx_norm)

        defo_norm = lut[:, 1].max()
        normalize(lut[:, 1], defo_norm)
        normalize(deform_4lut, defo_norm)

        # Perform interpolation
        emod = spint.griddata((lut[:, 0], lut[:, 1]), lut[:, 2],
                              (datax_4lut, deform_4lut),
                              method='linear')

        if extrapolate:
            # New in dclab 0.23.0: Perform extrapolation outside of the LUT
            # This is not well-tested and thus discouraged!
            extrapolate_emodulus(lut=lut,
                                 datax=datax_4lut,
                                 deform=deform_4lut,
                                 emod=emod,
                                 deform_norm=defo_norm,
                                 inplace=True)

        # Convert the LUT-interpolated emodulus back
        backscale_kw = {"channel_width_in": lut_meta["channel_width"],
                        "channel_width_out": channel_width,
                        "flow_rate_in": lut_meta["flow_rate"],
                        "flow_rate_out": flow_rate,
                        "viscosity_in": lut_meta["fluid_viscosity"],
                        "viscosity_out": visco,
                        "inplace": True}
        # deformation is not scaled (no units)
        scale_feature(feat=featx, data=datax_4lut, **backscale_kw)
        scale_emodulus(emod, **backscale_kw)
    else:
        # Corrections
        # We correct the LUT, because it contains less points than
        # the event data. Furthermore, the lut could be cached
        # in the future, if this takes up a lot of time.
        scale_kw = {"channel_width_in": lut_meta["channel_width"],
                    "channel_width_out": channel_width,
                    "flow_rate_in": lut_meta["flow_rate"],
                    "flow_rate_out": flow_rate,
                    "viscosity_in": lut_meta["fluid_viscosity"],
                    "viscosity_out": visco,
                    "inplace": True}
        # deformation is not scaled (no units)
        scale_feature(feat=featx, data=lut[:, 0], **scale_kw)
        scale_emodulus(lut[:, 2], **scale_kw)

        # Normalize interpolation data such that the spacing for
        # area and deformation is about the same during interpolation.
        featx_norm = lut[:, 0].max()
        normalize(lut[:, 0], featx_norm)
        normalize(datax, featx_norm)

        defo_norm = lut[:, 1].max()
        normalize(lut[:, 1], defo_norm)
        normalize(deform, defo_norm)

        # Perform interpolation
        emod = spint.griddata((lut[:, 0], lut[:, 1]), lut[:, 2],
                              (datax, deform),
                              method='linear')

        if extrapolate:
            # New in dclab 0.23.0: Perform extrapolation outside of the LUT
            # This is not well-tested and thus discouraged!
            extrapolate_emodulus(lut=lut,
                                 datax=datax,
                                 deform=deform,
                                 emod=emod,
                                 deform_norm=defo_norm,
                                 inplace=True)

    # Let the user know when the emodulus contains too many nan values
    nans = np.sum(np.isnan(emod))
    if nans / emod.size > 0.1:
        warnings.warn("The Young's modulus could not be computed for "
                      + "{:.0f}% of the data. ".format(nans/emod.size*100)
                      + "This is because they are not covered by the "
                      + "look-up table '{}'.".format(lut_data),
                      YoungsModulusLookupTableExceededWarning)

    return emod


def normalize(data, dmax):
    """Perform normalization in-place for interpolation

    Note that :func:`scipy.interpolate.griddata` has a `rescale`
    option which rescales the data onto the unit cube. For some
    reason this does not work well with LUT data, so we
    just normalize it by dividing by the maximum value.
    """
    assert isinstance(data, np.ndarray)
    data /= dmax
    return data



"""Test tdms file format"""
import os
import pathlib
import shutil
import sys
import tempfile

import numpy as np
import pytest

from dclab import new_dataset
import dclab.rtdc_dataset.fmt_tdms.naming

from helper_methods import retrieve_data


nptdms = pytest.importorskip("nptdms")


def test_compatibility_minimal():
    ds = new_dataset(retrieve_data("fmt-tdms_minimal_2016.zip"))
    assert ds["area_um"].ndim == 1  # important for matplotlib
    assert ds.config["setup"]["channel width"] == 20
    assert ds.config["setup"]["chip region"].lower() == "channel"
    assert ds.config["setup"]["flow rate"] == 0.12
    assert ds.config["imaging"]["pixel size"] == 0.34


@pytest.mark.skipif(sys.version_info < (3, 6),
                    reason="requires python3.6 or higher")
def test_compatibility_channel_width():
    # At some point, "Channel width" was replaced by "Channel width [um]"
    path = retrieve_data("fmt-tdms_minimal_2016.zip")
    para = path.parent / "M1_para.ini"
    pardata = para.read_text()
    pardata = pardata.replace("Channel width = 20\n", "Channel width = 34\n")
    para.write_text(pardata)
    ds = new_dataset(path)
    assert ds.config["setup"]["channel width"] == 34


def test_compatibility_shapein201():
    ds = new_dataset(retrieve_data("fmt-tdms_shapein-2.0.1-no-image_2017.zip"))
    assert ds.config["setup"]["channel width"] == 20
    assert ds.config["setup"]["chip region"].lower() == "channel"
    assert ds.config["setup"]["software version"] == "ShapeIn 2.0.1"
    assert ds.config["imaging"]["pixel size"] == 0.34
    assert ds.config["imaging"]["flash duration"] == 2
    assert ds.config["experiment"]["date"] == "2017-10-12"
    assert ds.config["experiment"]["time"] == "12:54:31"


@pytest.mark.filterwarnings(
    "ignore::dclab.rtdc_dataset.config.WrongConfigurationTypeWarning")
def test_contains_non_scalar():
    ds1 = new_dataset(retrieve_data("fmt-tdms_fl-image_2016.zip"))
    assert "contour" in ds1
    assert "image" in ds1
    assert "mask" in ds1
    assert "trace" in ds1
    ds2 = new_dataset(retrieve_data("fmt-tdms_minimal_2016.zip"))
    assert "image" not in ds2
    assert "trace" not in ds2
    ds3 = new_dataset(retrieve_data(
        "fmt-tdms_shapein-2.0.1-no-image_2017.zip"))
    assert "contour" not in ds3
    assert "image" not in ds3
    assert "mask" not in ds3
    assert "trace" not in ds3


@pytest.mark.filterwarnings(
    "ignore::dclab.rtdc_dataset.config.WrongConfigurationTypeWarning")
def test_contour_basic():
    ds = new_dataset(retrieve_data("fmt-tdms_fl-image_2016.zip"))
    assert len(ds["contour"]) == 12
    assert np.allclose(np.average(ds["contour"][0]), 38.488764044943821)
    assert ds["contour"]._initialized


@pytest.mark.filterwarnings(
    "ignore::dclab.rtdc_dataset.config.WrongConfigurationTypeWarning")
def test_contour_corrupt():
    path = pathlib.Path(retrieve_data("fmt-tdms_fl-image_2016.zip"))
    cpath = path.with_name("M1_contours.txt")
    # remove initial contours
    with cpath.open("r") as fd:
        lines = fd.readlines()
    with cpath.open("w") as fd:
        fd.writelines(lines[183:])
    ds = new_dataset(path)
    try:
        ds["contour"].determine_offset()
    except dclab.rtdc_dataset.fmt_tdms.event_contour.ContourIndexingError:
        pass
    else:
        assert False


def test_contour_unknown_offset():
    tdms_path = retrieve_data("fmt-tdms_fl-image-bright_2017.zip")
    cont_path = tdms_path.with_name("M4_0.040000ul_s_contours.txt")
    # Insert a false contour
    imaginary_contour = "\n".join([
        "Contour in frame 141",
        "(198, 30)",
        "(197, 29)",
        "(196, 28)",
        "(196, 27)",
        "(195, 27)",
        "(194, 26)",
        "(193, 25)",
        "(192, 24)",
        "(191, 25)",
        "",
    ])
    cont_data = cont_path.read_text()
    cont_path.write_text(imaginary_contour + cont_data)
    ds = dclab.new_dataset(tdms_path)
    with pytest.raises(dclab.rtdc_dataset.fmt_tdms.exc.ContourIndexingError,
                       match=r"Contour data has unknown offset \(frame 141\)"):
        ds["contour"].shape  # raises the error

    with pytest.raises(dclab.rtdc_dataset.fmt_tdms.exc.ContourIndexingError,
                       match=r"Contour data has unknown offset \(frame 141\)"):
        ds["contour"][0]  # raises the error


def test_contour_wrong_frame_number():
    tdms_path = retrieve_data("fmt-tdms_fl-image-bright_2017.zip")
    cont_path = tdms_path.with_name("M4_0.040000ul_s_contours.txt")
    # Insert a false contour
    contours = cont_path.read_text()
    contours = contours.replace("Contour in frame 1410 ",
                                "Contour in frame 1411 ")
    cont_path.write_text(contours)
    ds = dclab.new_dataset(tdms_path)
    # sanity check
    assert np.all(ds["contour"][1] != 0)
    # test for regression before 0.39.18
    assert ds["frame"][3] == 1410
    assert np.all(ds["contour"][3][0] == (190, 24))
    assert np.all(ds["contour"][3][2] == (192, 25))
    assert np.all(ds["contour"][3][-2] == (192, 24))
    assert np.all(ds["contour"][3][-1] == (191, 25))


def test_contour_naming():
    # Test that we always find the correct contour name
    ds = new_dataset(retrieve_data("fmt-tdms_minimal_2016.zip"))
    dp = pathlib.Path(ds.path).resolve()
    dn = dp.parent
    contfile = dn / "M1_0.120000ul_s_contours.txt"
    contfileshort = dn / "M1_contours.txt"
    contfileexact = dn / "M1_2us_70A_0.120000ul_s_contours.txt"
    del ds

    # Test for perfect match
    # "M1_2us_70A_0.120000ul_s_contours.txt" should have priority over
    # "M1_contours.txt" and "M1_0.120000ul_s_contours.txt".
    shutil.copy(str(contfile), str(contfileshort))
    shutil.copy(str(contfile), str(contfileexact))
    ds2 = new_dataset(dp)
    assert str(ds2["contour"].identifier) == str(contfileexact)
    assert not np.allclose(ds2["contour"][1], 0)
    del ds2

    # Check if "M1_contours.txt" is used if the other is not
    # there.
    contfileshort.unlink()
    contfileexact.unlink()
    contfile.rename(contfileshort)
    ds3 = new_dataset(dp)
    assert str(ds3["contour"].identifier) == str(contfileshort)
    del ds3
    contfileshort.rename(contfile)

    # Create M10 file
    with (dn / "M10_contours.txt").open(mode="w"):
        pass
    ds4 = new_dataset(dp)
    assert str(ds4["contour"].identifier) == str(contfile)
    del ds4

    # Check when there is no contour file
    os.remove(str(contfile))
    # This will issue a warning that no contour data was found.
    ds5 = new_dataset(dp)
    assert ds5["contour"].identifier is None


@pytest.mark.filterwarnings(
    "ignore::dclab.rtdc_dataset.config.WrongConfigurationTypeWarning")
def test_contour_negative_offset():
    ds = new_dataset(retrieve_data("fmt-tdms_fl-image_2016.zip"))
    ds["contour"][0]
    ds["contour"].event_offset = 1
    assert np.all(ds["contour"][0] == np.zeros((2, 2), dtype=int))


@pytest.mark.filterwarnings(
    "ignore::dclab.rtdc_dataset.config.WrongConfigurationTypeWarning")
def test_contour_not_initialized():
    ds = new_dataset(retrieve_data("fmt-tdms_fl-image_2016.zip"))
    assert not ds["contour"]._initialized


def test_tdms_shape_contour():
    ds = new_dataset(retrieve_data("fmt-tdms_fl-image-bright_2017.zip"))
    assert ds["contour"].shape == (8, np.nan, 2)
    assert len(ds["contour"]) == 8


def test_hdf5_shape_image():
    ds = new_dataset(retrieve_data("fmt-tdms_fl-image-bright_2017.zip"))
    assert "image" in ds.features_innate
    # Yeah, so the scalar features have different lengths than the
    # non-scalar features. This file is obviously broken (manually
    # cropped that data part), but it is sufficient for this test.
    assert len(ds) == 1008
    assert len(ds["image"]) == 9
    assert ds["image"].shape == (9, 80, 250)


def test_hdf5_shape_mask():
    ds = new_dataset(retrieve_data("fmt-tdms_fl-image-bright_2017.zip"))
    assert "mask" in ds.features_innate
    # Yeah, so the scalar features have different lengths than the
    # non-scalar features. This file is obviously broken (manually
    # cropped that data part), but it is sufficient for this test.
    assert len(ds) == 1008
    # mask length is determined by contour length
    assert len(ds["mask"]) == 8
    assert ds["mask"].shape == (8, 80, 250)


def test_hdf5_shape_trace():
    ds = new_dataset(retrieve_data("fmt-tdms_fl-image_2016.zip"))
    assert len(ds) == 44
    assert "trace" in ds.features_innate
    assert ds["trace"].shape == (2, 44, 1000)
    assert ds["trace"]["fl1_raw"].shape == (44, 1000)
    assert ds["trace"]["fl1_raw"][0].shape == (1000,)
    assert len(ds["trace"]) == 2
    assert len(ds["trace"]["fl1_raw"]) == 44
    assert len(ds["trace"]["fl1_raw"][0]) == 1000


def test_fluorescence_config():
    ds1 = new_dataset(retrieve_data("fmt-tdms_minimal_2016.zip"))
    assert "fluorescence" not in ds1.config
    ds2 = new_dataset(retrieve_data("fmt-tdms_2fl-no-image_2017.zip"))
    assert "fluorescence" in ds2.config


@pytest.mark.filterwarnings(
    "ignore::dclab.rtdc_dataset.config.WrongConfigurationTypeWarning")
@pytest.mark.filterwarnings('ignore::dclab.rtdc_dataset.'
                            + 'fmt_tdms.exc.'
                            + 'InitialFrameMissingWarning')
def test_image_basic():
    ds = new_dataset(retrieve_data("fmt-tdms_fl-image_2016.zip"))
    # Transition image
    assert np.allclose(ds["image"][0], 0)
    # Real image
    # 2025-01-23: paulmueller added the atol value, because CI was failing
    #             on macOS. Possible reason is underlying ffmpeg build
    #             for M1 etc. arm processors :zany_face:
    assert np.allclose(np.average(ds["image"][1]), 45.1490478515625,
                       atol=0.5)


@pytest.mark.filterwarnings(
    "ignore::dclab.rtdc_dataset.config.WrongConfigurationTypeWarning")
def test_image_column_length():
    ds = new_dataset(retrieve_data("fmt-tdms_fl-image_2016.zip"))
    assert len(ds["image"]) == 3


@pytest.mark.filterwarnings(
    "ignore::dclab.rtdc_dataset.config.WrongConfigurationTypeWarning")
def test_image_corrupt():
    path = pathlib.Path(retrieve_data("fmt-tdms_fl-image_2016.zip"))
    vpath = path.with_name("M1_imaq.avi")
    # create empty video file
    vpath.unlink()
    vpath.touch()
    try:
        new_dataset(path)
    except dclab.rtdc_dataset.fmt_tdms.exc.InvalidVideoFileError:
        pass
    else:
        assert False


@pytest.mark.filterwarnings(
    "ignore::dclab.rtdc_dataset.config.WrongConfigurationTypeWarning")
@pytest.mark.filterwarnings('ignore::dclab.rtdc_dataset.'
                            + 'fmt_tdms.exc.CorruptFrameWarning')
def test_image_out_of_bounds():
    ds = new_dataset(retrieve_data("fmt-tdms_fl-image_2016.zip"))
    assert len(ds["image"]) == 3
    assert np.allclose(ds["image"][0], 0)  # dummy
    assert not np.allclose(ds["image"][1], 0)
    assert not np.allclose(ds["image"][2], 0)
    assert np.allclose(ds["image"][3], 0)  # causes warning


@pytest.mark.parametrize("feat", ["image", "mask"])
@pytest.mark.parametrize("idxs", [slice(0, 3), np.arange(3),
                                  [0, 1, 2], [True, True, True, False]])
def test_index_slicing_tdms_fails(feat, idxs):
    """The tdms-file format does not support slice/array indexing"""
    data = retrieve_data("fmt-tdms_shapein-2.0.1-no-image_2017.zip")
    ds = new_dataset(data)

    with pytest.raises(NotImplementedError, match="scalar integers"):
        ds[feat][idxs]


@pytest.mark.filterwarnings('ignore::dclab.rtdc_dataset.'
                            + 'fmt_tdms.exc.'
                            + 'InitialFrameMissingWarning')
@pytest.mark.filterwarnings('ignore::dclab.rtdc_dataset.'
                            + 'feat_anc_core.ancillary_feature.'
                            + 'BadFeatureSizeWarning')
def test_large_fov():
    ds = new_dataset(retrieve_data("fmt-tdms_fl-image-large-fov_2017.zip"))
    # initial image is missing
    assert np.allclose(ds["image"][0], 0)
    # initial contour is empty
    assert np.allclose(ds["contour"][0], 0)
    # maximum of contour is larger than 255 (issue #167)
    assert ds["contour"][1].max() == 815
    # compute brightness with given contour
    # Remove the brightness column and let it recompute
    # using the ancillary columns. Besides testing the
    # correct positioning of the contour, this is a
    # sanity test for the brightness computation.
    bavg = ds._events.pop("bright_avg")
    bcom = ds["bright_avg"]
    assert np.allclose(bavg[1], bcom[1])


@pytest.mark.filterwarnings(
    "ignore::dclab.rtdc_dataset.config.WrongConfigurationTypeWarning")
@pytest.mark.parametrize("exname", [
    "fmt-tdms_minimal_2016.zip",
    "fmt-tdms_fl-image_2016.zip",
    "fmt-tdms_fl-image-bright_2017.zip",
    "fmt-tdms_fl-image-large-fov_2017.zip",
    "fmt-tdms_shapein-2.0.1-no-image_2017.zip"])
def test_load_tdms_all(exname):
    tdms_path = retrieve_data(exname)
    new_dataset(tdms_path)


@pytest.mark.filterwarnings(
    "ignore::dclab.rtdc_dataset.config.WrongConfigurationTypeWarning")
def test_load_tdms_avi_files_1():
    tdms_path = retrieve_data("fmt-tdms_fl-image_2016.zip")
    edest = pathlib.Path(tdms_path).parent
    with new_dataset(tdms_path) as ds1:
        assert pathlib.Path(ds1["image"].video_file).name == "M1_imaq.avi"
    shutil.copyfile(str(edest / "M1_imaq.avi"),
                    str(edest / "M1_imag.avi"))
    with new_dataset(tdms_path) as ds2:
        # prefer imag over imaq
        assert pathlib.Path(ds2["image"].video_file).name == "M1_imag.avi"
    shutil.copyfile(str(edest / "M1_imaq.avi"),
                    str(edest / "M1_test.avi"))
    with new_dataset(tdms_path) as ds3:
        # ignore any other videos
        assert pathlib.Path(ds3["image"].video_file).name == "M1_imag.avi"


@pytest.mark.filterwarnings(
    "ignore::dclab.rtdc_dataset.config.WrongConfigurationTypeWarning")
def test_load_tdms_avi_files_2():
    tdms_path = retrieve_data("fmt-tdms_fl-image_2016.zip")
    edest = pathlib.Path(tdms_path).parent
    shutil.copyfile(str(edest / "M1_imaq.avi"),
                    str(edest / "M1_test.avi"))
    os.remove(str(edest / "M1_imaq.avi"))
    with new_dataset(tdms_path) as ds4:
        # use available video if ima* not there
        assert pathlib.Path(ds4["image"].video_file).name == "M1_test.avi"


def test_load_tdms_simple():
    tdms_path = retrieve_data("fmt-tdms_minimal_2016.zip")
    ds = new_dataset(tdms_path)
    assert ds.filter.all.shape[0] == 156


@pytest.mark.filterwarnings(
    "ignore::dclab.rtdc_dataset.config.WrongConfigurationTypeWarning")
@pytest.mark.filterwarnings('ignore::dclab.rtdc_dataset.'
                            + 'fmt_tdms.exc.'
                            + 'InitialFrameMissingWarning')
@pytest.mark.filterwarnings('ignore::dclab.rtdc_dataset.'
                            + 'feat_anc_core.ancillary_feature.'
                            + 'BadFeatureSizeWarning')
def test_mask_basic():
    ds = new_dataset(retrieve_data("fmt-tdms_fl-image_2016.zip"))
    assert len(ds["mask"]) == 12
    # Test mask computation by averaging brightness and comparing to
    # the ancillary feature "bright_avg".
    bavg1 = ds["bright_avg"][1]
    bavg2 = np.mean(ds["image"][1][ds["mask"][1]])
    assert np.allclose(bavg1, bavg2), "mask is correctly computed from contour"


@pytest.mark.filterwarnings(
    "ignore::dclab.rtdc_dataset.config.WrongConfigurationTypeWarning")
def test_mask_img_shape_1():
    path = retrieve_data("fmt-tdms_fl-image_2016.zip")
    # shape from image data
    with new_dataset(path) as ds:
        assert ds["mask"]._img_shape == (96, 256)


@pytest.mark.filterwarnings(
    "ignore::dclab.rtdc_dataset.config.WrongConfigurationTypeWarning")
def test_mask_img_shape_2():
    path = retrieve_data("fmt-tdms_fl-image_2016.zip")
    path.with_name("M1_imaq.avi").unlink()
    with new_dataset(path) as ds:
        # shape from config ("roi size x", "roi size y")
        assert ds["mask"]._img_shape == (96, 256)
    # no shape available
    with new_dataset(path) as ds:
        ds.config["imaging"].pop("roi size x")
        ds.config["imaging"].pop("roi size y")
        assert ds["mask"]._img_shape == (0, 0)
        assert len(ds["mask"]) == 0


@pytest.mark.filterwarnings(
    "ignore::dclab.rtdc_dataset.config.WrongConfigurationTypeWarning")
def test_mask_img_wrong_config_shape_1():
    path = retrieve_data("fmt-tdms_fl-image_2016.zip")
    with new_dataset(path) as ds:
        # deliberately set wrong size in ROI (fmt_tdms tries image shape first)
        ds.config["imaging"]["roi size x"] = 200
        ds.config["imaging"]["roi size y"] = 200
        assert ds["mask"]._img_shape == (96, 256)


@pytest.mark.filterwarnings(
    "ignore::dclab.rtdc_dataset.config.WrongConfigurationTypeWarning")
def test_mask_img_wrong_config_shape_2():
    path = retrieve_data("fmt-tdms_fl-image_2016.zip")
    path.with_name("M1_imaq.avi").unlink()
    with new_dataset(path) as ds:
        # deliberately set wrong size in ROI (fmt_tdms tries image shape first)
        ds.config["imaging"]["roi size x"] = 200
        ds.config["imaging"]["roi size y"] = 200
        assert ds["mask"]._img_shape == (200, 200)


def test_naming_valid():
    for key in dclab.rtdc_dataset.fmt_tdms.naming.dclab2tdms:
        assert dclab.definitions.feature_exists(key)


@pytest.mark.filterwarnings(
    "ignore::dclab.rtdc_dataset.config.WrongConfigurationTypeWarning")
def test_parameters_txt():
    ds = new_dataset(retrieve_data("fmt-tdms_fl_2015.zip"))
    assert ds.config["setup"]["module composition"] == "Cell_Flow_2, Fluor"
    assert ds.config["setup"]["software version"] == "fRT-DC V0.80 150601"
    assert ds.config["setup"]["identifier"] == "47 red angels"
    assert ds.config["setup"]["medium"] == "other"
    assert ds.config["setup"]["flow rate sample"] == 0.01
    assert ds.config["setup"]["flow rate sheath"] == 0.03
    assert ds.config["imaging"]["pixel size"] == 0.34
    assert ds.config["imaging"]["flash duration"] == 2
    assert ds.config["fluorescence"]["samples per event"] == 1000
    assert ds.config["fluorescence"]["sample rate"] == 1000000
    assert ds.config["fluorescence"]["signal min"] == -1
    assert ds.config["fluorescence"]["signal max"] == 1
    assert ds.config["fluorescence"]["trace median"] == 21
    assert ds.config["fluorescence"]["laser 1 power"] == 5
    assert ds.config["fluorescence"]["laser 2 power"] == 0
    assert ds.config["fluorescence"]["laser 3 power"] == 0


@pytest.mark.filterwarnings(
    "ignore::dclab.rtdc_dataset.config.WrongConfigurationTypeWarning")
def test_para_ini_fluorescence():
    """Read commented-out parameters from para.ini"""
    path = retrieve_data("fmt-tdms_fl_2015.zip")
    (path.parent / "parameters.txt").unlink()
    ds = new_dataset(path)

    assert ds.config["setup"]["module composition"] == "Cell_Flow_2, Fluor"
    assert ds.config["setup"]["software version"] == "fRT-DC V0.80 150601"
    assert ds.config["setup"]["identifier"] == "47 red angels"
    assert ds.config["setup"]["medium"] == "other"
    assert ds.config["setup"]["flow rate sample"] == 0.01
    assert ds.config["setup"]["flow rate sheath"] == 0.03
    assert ds.config["imaging"]["pixel size"] == 0.34
    assert ds.config["imaging"]["flash duration"] == 2
    assert ds.config["fluorescence"]["sample rate"] == 1000000
    assert ds.config["fluorescence"]["signal min"] == -1
    assert ds.config["fluorescence"]["signal max"] == 1


def test_pixel_size():
    path = retrieve_data("fmt-tdms_minimal_2016.zip")
    para = path.parent / "M1_para.ini"
    data = para.open("r").read()
    newdata = data.replace("Pix Size = 0.340000", "Pix Size = 0.120000")
    with para.open("w") as fd:
        fd.write(newdata)
    ds = new_dataset(path)
    assert ds.config["imaging"]["pixel size"] == 0.12


def test_project_path():
    tfile = retrieve_data("fmt-tdms_minimal_2016.zip")
    ds = dclab.new_dataset(tfile)
    assert ds.hash == "69733e31b005c145997fac8a22107ded"
    assert ds.format == "tdms"
    tpath = pathlib.Path(tfile).resolve()
    a = dclab.rtdc_dataset.fmt_tdms.get_project_name_from_path(str(tpath))
    b = dclab.rtdc_dataset.fmt_tdms.get_project_name_from_path(
        str(tpath.parent))
    assert a == b
    c = dclab.rtdc_dataset.fmt_tdms.get_project_name_from_path(
        str(tpath.parent / "online" / tpath.name))
    d = dclab.rtdc_dataset.fmt_tdms.get_project_name_from_path(
        str(tpath.parent / "online" / "data" / tpath.name))
    e = dclab.rtdc_dataset.fmt_tdms.get_project_name_from_path(
        str(tpath.parent / "online" / "data"))

    assert a == e
    assert a == c
    assert a == d


@pytest.mark.filterwarnings(
    "ignore::dclab.rtdc_dataset.config.WrongConfigurationTypeWarning")
def test_trace_basic():
    ds = new_dataset(retrieve_data("fmt-tdms_fl-image_2016.zip"))
    msg = "traces should not be loaded into memory before first access"
    assert ds["trace"].__repr__().count("<not loaded into memory>"), msg
    assert len(ds["trace"]) == 2
    assert np.allclose(np.average(
        ds["trace"]["fl1_median"][0]), 287.08999999999997)


@pytest.mark.filterwarnings(
    "ignore::dclab.rtdc_dataset.config.WrongConfigurationTypeWarning")
def test_trace_import_fail():
    # make sure undefined trace data does not raise an error
    tdms_path = retrieve_data("fmt-tdms_fl-image_2016.zip")
    dclab.definitions.FLUOR_TRACES.append("peter")
    dclab.rtdc_dataset.fmt_tdms.naming.tr_data_map["peter"] = [u'ukwn', u'ha']
    new_dataset(tdms_path)
    # clean up
    dclab.rtdc_dataset.fmt_tdms.naming.tr_data_map.pop("peter")
    dclab.definitions.FLUOR_TRACES.pop(-1)


@pytest.mark.filterwarnings(
    "ignore::dclab.rtdc_dataset.config.WrongConfigurationTypeWarning")
def test_trace_methods():
    ds = new_dataset(retrieve_data("fmt-tdms_fl-image_2016.zip"))
    for k in list(ds["trace"].keys()):
        assert k in dclab.definitions.FLUOR_TRACES
    for k in ds["trace"]:
        assert k in dclab.definitions.FLUOR_TRACES
    assert ds["trace"].__repr__().count("<loaded into memory>")


@pytest.mark.filterwarnings(
    "ignore::dclab.rtdc_dataset.config.WrongConfigurationTypeWarning")
@pytest.mark.filterwarnings('ignore::dclab.rtdc_dataset.'
                            + 'fmt_tdms.exc.'
                            + 'MultipleSamplesPerEventFound')  # desired
def test_trace_wrong_samples_per_event():
    """The "samples per event" should be a constant for trace data

    If this condition is not given, the trace data is considered
    unusable, but the rest of the data still makes sense
    (Philipp said that something must have gone wrong during writing
    of the trace data).
    """
    tdms = retrieve_data("fmt-tdms_fl-image_2016.zip")
    mdata = nptdms.TdmsFile(str(tdms))

    channels = []
    # recreate channels, because the data type might not be correct
    for ch in mdata.groups()[0].channels():
        if ch.data.size == 0:
            # not supported by nptdms 0.25.0
            continue
        channels.append(nptdms.ChannelObject("Cell Track", ch.name, ch.data))

    # blank write same data to test that modification works
    with nptdms.TdmsWriter(str(tdms)) as tdms_writer:
        tdms_writer.write_segment(channels)

    with dclab.new_dataset(tdms) as ds:
        assert "trace" in ds

    # modify objects
    sampleids = mdata["Cell Track"]["FL1index"].data
    sampleids[0] = 10
    sampleids[1] = 20
    sampleids[2] = 40
    wchannels = []
    for ch in channels:
        if ch.channel == "FL1index":
            nptdms.ChannelObject("Cell Track", "FL1index", sampleids)
        wchannels.append(ch)

    with nptdms.TdmsWriter(str(tdms)) as tdms_writer:
        tdms_writer.write_segment(wchannels)

    with dclab.new_dataset(tdms) as ds:
        assert "trace" not in ds


@pytest.mark.filterwarnings(
    "ignore::dclab.rtdc_dataset.config.WrongConfigurationTypeWarning")
def test_unicode_paths():
    path = retrieve_data("fmt-tdms_fl-image_2016.zip")
    path = pathlib.Path(path)
    pp = path.parent
    # create a unicode name
    pp2 = pathlib.Path(tempfile.mktemp(prefix="dclàb_tést_asgård_únícodè"))
    pp.rename(pp2)
    ds = new_dataset(pp2 / path.name)
    ds.__repr__()
    shutil.rmtree(str(pp2), ignore_errors=True)


if __name__ == "__main__":
    # Run all tests
    loc = locals()
    for key in list(loc.keys()):
        if key.startswith("test_") and hasattr(loc[key], "__call__"):
            loc[key]()



import os
from os.path import join
import tempfile

import pytest

import dclab
from dclab import new_dataset

from helper_methods import example_data_dict, retrieve_data


@pytest.mark.filterwarnings(
    "ignore::dclab.rtdc_dataset.config.WrongConfigurationTypeWarning")
@pytest.mark.filterwarnings('ignore::dclab.rtdc_dataset.'
                            + 'fmt_tdms.exc.CorruptFrameWarning')
def test_avi_export():
    pytest.importorskip("nptdms")
    ds = new_dataset(retrieve_data("fmt-tdms_fl-image_2016.zip"))
    edest = tempfile.mkdtemp()
    f1 = join(edest, "test.avi")
    ds.export.avi(path=f1)
    assert os.stat(
        f1)[6] > 1e4, "Resulting file to small, Something went wrong!"


@pytest.mark.filterwarnings(
    "ignore::dclab.rtdc_dataset.config.WrongConfigurationTypeWarning")
@pytest.mark.filterwarnings('ignore::dclab.rtdc_dataset.'
                            + 'fmt_tdms.exc.CorruptFrameWarning')
def test_avi_override():
    pytest.importorskip("nptdms")
    ds = new_dataset(retrieve_data("fmt-tdms_fl-image_2016.zip"))

    edest = tempfile.mkdtemp()
    f1 = join(edest, "test.avi")
    ds.export.avi(f1, override=True)
    try:
        ds.export.avi(f1[:-4], override=False)
    except OSError:
        pass
    else:
        raise ValueError("Should append .avi and not override!")


def test_avi_no_images():
    pytest.importorskip("imageio")
    keys = ["area_um", "deform", "time", "frame", "fl3_width"]
    ddict = example_data_dict(size=127, keys=keys)
    ds = dclab.new_dataset(ddict)

    edest = tempfile.mkdtemp()
    f1 = join(edest, "test.avi")
    try:
        ds.export.avi(f1)
    except OSError:
        pass
    else:
        raise ValueError("There should be no image data to write!")



import numpy as np
from scipy.optimize import fsolve


def find_upper_intersection(func, points, xrange):
    options = []
    for ii in range(len(points)):
        x1, y1 = points[ii]
        x2, y2 = points[ii-1]
        def funcs(x): return y1 + (y2-y1)/(x2-x1)*(x-x1)
        ret = fsolve(lambda x: func(x) - funcs(x), xrange[0])
        for r in ret:
            if xrange[0] < r < xrange[1]:
                options.append([r, func(r)])
    options = np.array(options)
    # If there are multiple options, choose the one that is closest to
    # xrange[0].
    ido = np.argmin(np.abs(options[:, 0] - xrange[0]))
    return options[ido]



import numpy as np
import pytest

import dclab
from dclab.rtdc_dataset.feat_anc_core.ancillary_feature import (
    AncillaryFeature, BadFeatureSizeWarning)

from helper_methods import example_data_dict, retrieve_data, calltracker


@pytest.fixture
def fake_af_cleanup_fixture():
    """Fixture used to set up and cleanup some fake ancillary features"""
    # code run before the test
    # set the method calls of our shared method to zero (just in case)
    shared_fake_af_method.calls = 0
    # setup fake ancillary features
    af1, af2 = setup_fake_af()
    # variables yielded (input) to the test, then the test is run
    yield af1, af2
    # code run after the test
    # set the method calls of our shared method to zero
    shared_fake_af_method.calls = 0
    # remove our fake ancillary feature instances from the registry
    AncillaryFeature.features.remove(af1)
    AncillaryFeature.features.remove(af2)
    AncillaryFeature.feature_names.remove(af1.feature_name)
    AncillaryFeature.feature_names.remove(af2.feature_name)


@calltracker
def shared_fake_af_method(rtdc_ds):
    """An example of an `AncillaryFeature.method` that is used to calculate two
    different ancillary features.
    """
    cw = rtdc_ds.config["setup"]["channel width"]
    data_dict = {
        "userdef1": np.arange(1, len(rtdc_ds) + 1) * cw,
        "userdef2": np.arange(1, len(rtdc_ds) + 1) * cw * 2
    }
    return data_dict


def setup_fake_af():
    """Creates some example ancillary features"""
    af1 = AncillaryFeature(feature_name="userdef1",
                           method=shared_fake_af_method,
                           req_config=[["setup", ["channel width"]]])
    af2 = AncillaryFeature(feature_name="userdef2",
                           method=shared_fake_af_method,
                           req_config=[["setup", ["channel width"]]])
    return af1, af2


@pytest.mark.filterwarnings(
    "ignore::dclab.rtdc_dataset.config.WrongConfigurationTypeWarning")
def test_af_0basic():
    pytest.importorskip("nptdms")
    ds = dclab.new_dataset(retrieve_data("fmt-tdms_fl-image_2016.zip"))
    for cc in ['fl1_pos',
               'frame',
               'size_x',
               'size_y',
               'contour',
               'area_cvx',
               'circ',
               'image',
               'trace',
               'fl1_width',
               'nevents',
               'pos_x',
               'pos_y',
               'fl1_area',
               'fl1_max',
               ]:
        assert cc in ds

    # ancillaries
    for cc in ["deform",
               "area_um",
               "aspect",
               "frame",
               "index",
               "time",
               ]:
        assert cc in ds


def test_af_0error():
    keys = ["circ"]
    ddict = example_data_dict(size=8472, keys=keys)
    ds = dclab.new_dataset(ddict)
    try:
        ds["unknown_column"]
    except KeyError:
        pass
    else:
        raise ValueError("Should have raised KeyError!")


def test_af_aspect():
    pytest.importorskip("nptdms")
    # Aspect ratio of the data
    ds = dclab.new_dataset(retrieve_data("fmt-tdms_fl-image-bright_2017.zip"))
    aspect = ds["aspect"]
    assert np.sum(aspect > 1) == 904
    assert np.sum(aspect < 1) == 48


@pytest.mark.filterwarnings(
    "ignore::dclab.rtdc_dataset.config.WrongConfigurationTypeWarning")
def test_af_area_ratio():
    pytest.importorskip("nptdms")
    ds = dclab.new_dataset(retrieve_data("fmt-tdms_fl-image_2016.zip"))
    comp_ratio = ds["area_ratio"]
    # The convex area is always >= the raw area
    assert np.all(comp_ratio >= 1)
    assert np.allclose(comp_ratio[0], 1.0196464)


def test_af_deform():
    keys = ["circ"]
    ddict = example_data_dict(size=8472, keys=keys)
    ds = dclab.new_dataset(ddict)
    assert np.allclose(ds["deform"], 1 - ds["circ"])


@pytest.mark.filterwarnings(
    "ignore::dclab.rtdc_dataset.config.WrongConfigurationTypeWarning")
@pytest.mark.parametrize("path", [
    "fmt-tdms_minimal_2016.zip",
    "fmt-tdms_fl-image_2016.zip",
    "fmt-tdms_fl-image-bright_2017.zip",
    "fmt-tdms_fl-image-large-fov_2017.zip",
    "fmt-tdms_shapein-2.0.1-no-image_2017.zip",
])
def test_af_method_called_once_with_shared_pipeline(
        fake_af_cleanup_fixture, path):
    """Verifies that when an `AncillaryFeature.method` is shared between
    ancillary features, the `method` itself is only called once. All ancillary
    features that share the `method` will be populated in this single call.
    """
    pytest.importorskip("nptdms")
    af1, af2 = fake_af_cleanup_fixture
    ds = dclab.new_dataset(retrieve_data(path))
    assert af1.feature_name not in ds.features_innate
    assert af2.feature_name not in ds.features_innate
    assert af1.feature_name in ds
    assert af2.feature_name in ds

    # handle tdms "time" `AncillaryFeature`
    tdms_time_feature = False
    feats = [af1.feature_name, af2.feature_name]
    if ds.format == "tdms":
        tdms_time_feature = True
        feats.append("time")

    assert shared_fake_af_method.calls == 0
    assert len(ds._ancillaries) == 0 + tdms_time_feature

    _ = ds[af1.feature_name]
    assert shared_fake_af_method.calls == 1
    assert len(ds._ancillaries) == 2 + tdms_time_feature
    assert all(k in ds._ancillaries for k in feats)

    _ = ds[af2.feature_name]
    assert shared_fake_af_method.calls == 1, "method is not called again"
    assert len(ds._ancillaries) == 2 + tdms_time_feature


@pytest.mark.filterwarnings(
    "ignore::dclab.rtdc_dataset.config.WrongConfigurationTypeWarning")
@pytest.mark.parametrize("path", [
    "fmt-tdms_minimal_2016.zip",
    "fmt-tdms_fl-image_2016.zip",
    "fmt-tdms_fl-image-bright_2017.zip",
    "fmt-tdms_fl-image-large-fov_2017.zip",
    "fmt-tdms_shapein-2.0.1-no-image_2017.zip",
])
def test_af_recomputed_on_hash_change(fake_af_cleanup_fixture, path):
    """Check whether features are recomputed when the hash changes"""
    pytest.importorskip("nptdms")
    af1, af2 = fake_af_cleanup_fixture
    ds = dclab.new_dataset(retrieve_data(path))
    assert af1.feature_name not in ds.features_innate
    assert af2.feature_name not in ds.features_innate
    assert af1.feature_name in ds
    assert af2.feature_name in ds

    ud1a = ds[af1.feature_name]
    ud2a = ds[af2.feature_name]

    ds.config["setup"]["channel width"] *= 1.1

    ud1b = ds[af1.feature_name]
    ud2b = ds[af2.feature_name]

    assert np.all(ud1a != ud1b)
    assert np.allclose(ud1a * 1.1, ud1b)
    assert np.all(ud2a != ud2b)
    assert np.allclose(ud2a * 1.1, ud2b)


def test_af_time():
    pytest.importorskip("nptdms")
    ds = dclab.new_dataset(retrieve_data("fmt-tdms_minimal_2016.zip"))
    time = ds["time"]
    frame = ds["frame"]
    fr = ds.config["imaging"]["frame rate"]
    assert np.allclose(time[0], frame[0] / fr, atol=0, rtol=1e-7)
    assert np.allclose(time[1], 82.526, atol=0, rtol=1e-7)
    assert np.all(np.diff(time) > 0)


@pytest.mark.filterwarnings(
    "ignore::dclab.rtdc_dataset.config.WrongConfigurationTypeWarning")
def test_af_warning_from_check_data_size():
    h5path = retrieve_data("fmt-hdf5_fl_2018.zip")
    with dclab.new_dataset(h5path) as ds:
        data = np.arange(len(ds)//2)
        data_dict = {"name1": data}
        with pytest.warns(BadFeatureSizeWarning):
            AncillaryFeature.check_data_size(ds, data_dict)


def test_af_feature_property_core():
    ddict = example_data_dict(size=67, keys=["area_um", "circ"])
    ds = dclab.new_dataset(ddict)
    assert "deform" in ds.features_ancillary
    assert ds.features_ancillary == ["deform", "index"]
    assert ds.features == ["area_um", "circ", "deform", "index"]
    # trigger computation of deformation
    assert np.allclose(ds["circ"], 1 - ds["deform"])
    # check again
    assert ds.features_ancillary == ["deform", "index"]


def test_af_feature_property_core_control():
    # Here, deformation is already an innate feature
    ddict = example_data_dict(size=67, keys=["area_um", "circ", "deform"])
    ds = dclab.new_dataset(ddict)
    assert "deform" not in ds.features_ancillary
    assert ds.features_ancillary == ["index"]
    assert ds.features == ["area_um", "circ", "deform", "index"]
